# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from typing import Tuple, List, Union, Optional

from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup, RunnersGroupResult, CollectorCallback, AbstractGroup, RunResult

predefined_file_path = "/etc/modules-load.d/predefined.conf"

ERROR_BLANK_MODULE = "Found blank kernel module at path {path}"
ERROR_DUPLICATE_MODULE = "Kernel module {module_name!r} is duplicated in the inventory"


@enrichment(EnrichmentStage.FULL)
def enrich_kernel_modules(cluster: KubernetesCluster) -> None:
    """
    The method enrich the list of kernel modules ('services.modprobe') according to OS family
    """

    for os_family in ('debian', 'rhel', 'rhel8', 'rhel9'):
        # Remove the section for OS families if no node has these OS families.
        modprobe_config = cluster.inventory["services"]["modprobe"]
        if cluster.nodes['all'].get_subgroup_with_os(os_family).is_empty():
            del modprobe_config[os_family]
            continue

        modprobe_config[os_family] = modules_list = _convert_modprobe_config(cluster, os_family)

        _verify_modules_list(modules_list)
        _apply_defaults(cluster, modules_list)


def _convert_modprobe_config(cluster: KubernetesCluster, os_family: str) -> List[dict]:
    modprobe_config: List[Union[str, dict]] = cluster.inventory["services"]["modprobe"][os_family]

    modules_list: List[dict] = []
    for i, module_name in enumerate(modprobe_config):
        config = _convert_module(module_name, ["services", "modprobe", os_family, i])
        if config is not None:
            modules_list.append(config)

    return modules_list


def _convert_module(module_name: Union[str, dict], path: List[Union[str, int]]) -> Optional[dict]:
    if isinstance(module_name, str):
        # Obsolete approach to render values. Empty modules should be removed from the list.
        module_name = module_name.strip()
        if module_name == '':
            return None

        config = {
            'modulename': module_name,
        }
    else:
        config = module_name

    config['modulename'] = config['modulename'].strip()
    if config['modulename'] == '':
        raise Exception(ERROR_BLANK_MODULE.format(path=utils.pretty_path(path + ['modulename'])))

    return config


def _verify_modules_list(modules_list: List[dict]) -> None:
    known_modules = set()
    for config in modules_list:
        module_name = config['modulename']
        if module_name in known_modules:
            raise Exception(ERROR_DUPLICATE_MODULE.format(module_name=module_name))

        known_modules.add(module_name)


def _apply_defaults(cluster: KubernetesCluster, modules_list: List[dict]) -> None:
    for config in modules_list:
        if config.get('groups') is None and config.get('nodes') is None:
            config['groups'] = ['control-plane', 'worker', 'balancer']

        config.setdefault('install', True)

        if config.get('nodes') is not None:
            all_nodes_names = cluster.nodes['all'].get_nodes_names()
            unknown_nodes = set(config['nodes']) - set(all_nodes_names)
            if unknown_nodes:
                # Only warn instead of raising an error to allow remove & add the same node.
                cluster.log.warning(
                    f"Unknown node names {', '.join(map(repr, unknown_nodes))} "
                    f"provided for kernel module {config['modulename']!r}. ")


def generate_config(node: AbstractGroup[RunResult]) -> str:
    cluster: KubernetesCluster = node.cluster
    config = ''
    modprobe_config: List[dict] = cluster.inventory['services']['modprobe'][node.get_nodes_os()]
    for module_config in modprobe_config:
        group = cluster.create_group_from_groups_nodes_names(
            module_config.get('groups', []), module_config.get('nodes', []))

        if not module_config['install'] or not group.has_node(node.get_node_name()):
            continue

        config += module_config['modulename'] + "\n"

    return config


def setup_modprobe(group: NodeGroup) -> bool:
    cluster: KubernetesCluster = group.cluster
    logger = cluster.log

    is_valid, is_config_valid, result = is_modprobe_valid(group)

    if is_valid and is_config_valid:
        logger.debug("Skipped - all necessary kernel modules are presented")
        logger.debug(result)
        return False

    defer = group.new_defer()
    for node in defer.get_ordered_members_list():
        config = generate_config(node)
        if not config:
            continue
        raw_config = config.replace('\n', ' ')

        logger.debug("Uploading config...")
        dump_filename = f'modprobe/modprobe_predefined_{node.get_node_name()}.conf'

        utils.dump_file(cluster, config, dump_filename)
        node.put(io.StringIO(config), predefined_file_path, backup=True, sudo=True)
        node.sudo("modprobe -a %s" % raw_config)

    defer.flush()

    return True


def is_modprobe_valid(group: NodeGroup) -> Tuple[bool, bool, RunnersGroupResult]:
    cluster: KubernetesCluster = group.cluster
    logger = cluster.log
    defer = group.new_defer()

    lsmod_collector = CollectorCallback(cluster)
    config_collector = CollectorCallback(cluster)
    defer.sudo("lsmod", warn=True, callback=lsmod_collector)
    defer.sudo(f"cat {predefined_file_path}", warn=True, callback=config_collector)
    defer.flush()

    is_valid = True
    is_config_valid = True

    for node in defer.get_ordered_members_list():
        expected_config = generate_config(node)
        if not expected_config:
            continue

        expected_modules = expected_config.rstrip('\n').split('\n')

        host = node.get_host()

        lsmod_result = lsmod_collector.result[host]
        actual_modules = {mod.split()[0]
                          for mod in lsmod_result.stdout.rstrip('\n').split('\n')[1:]}

        actual_config = config_collector.result[host]

        for module_name in expected_modules:
            if module_name not in actual_modules:
                logger.debug(f'Kernel module {module_name} is not found at {host}')
                is_valid = False

        if expected_config != actual_config.stdout:
            logger.debug(f'Config is outdated at {host}')
            is_config_valid = False

    return is_valid, is_config_valid, lsmod_collector.result



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import io
import random
import time
from typing import Optional, List

from jinja2 import Template

from kubemarine import system, packages
from kubemarine.core import utils, static
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup, NodeConfig, RunnersGroupResult, CollectorCallback, DeferredGroup


def autodetect_interface(cluster: KubernetesCluster, name: str) -> str:
    address = cluster.nodes['all'].get_member_by_name(name).get_host()
    interface: str = cluster.nodes_context.get(address, {}).get('active_interface')
    # If undefined, still return it. The error about inaccessible nodes will be raised later.
    if interface:
        return interface

    raise Exception('Failed to autodetect active interface for %s' % name)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['add_node'])
def enrich_add_node_vrrp_ips(_: KubernetesCluster) -> None:
    # If "vrrp_ips" section is ever supported when adding node,
    # It will be necessary to more accurately install and reconfigure the keepalived on existing nodes.

    # if "vrrp_ips" in cluster.procedure_inventory:
    #     utils.merge_vrrp_ips(cluster.procedure_inventory, inventory)
    pass


@enrichment(EnrichmentStage.PROCEDURE, procedures=['remove_node'])
def enrich_remove_node_vrrp_ips(_: KubernetesCluster) -> None:
    # Do not remove VRRP IPs and do not change their assigned hosts.
    # If the assigned host does not exist or is not a balancer, it will be just skipped.
    pass


@enrichment(EnrichmentStage.FULL)
def enrich_inventory_apply_defaults(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory

    # if vrrp_ips is empty, then nothing to do
    if not inventory['vrrp_ips']:
        return

    logger = cluster.log

    balancers = get_all_balancer_names(inventory)

    logger.verbose("Detected default keepalived hosts: %s" % balancers)
    if not balancers:
        logger.warning("VRRP IPs are specified, but there are no balancers. Keepalived will not be configured.")

    # iterate over each vrrp_ips item and check if any hosts defined to be used in it
    for i, item in enumerate(inventory['vrrp_ips']):

        if isinstance(item, str):
            inventory['vrrp_ips'][i] = item = {
                'ip': item
            }

        # is router_id defined?
        if item.get('router_id') is None:
            # is there ipv6?
            if ':' in item['ip']:
                item['router_id'] = item['ip'].split(':').pop()
                if item['router_id'] == '':
                    item['router_id'] = '0'
                # in adress with long last octet e.g. "765d" it is necessary to use only last "5d" and convert it from hex to int
                item['router_id'] = str(int(item['router_id'][-2:], 16))
            else:
                item['router_id'] = item['ip'].split('.').pop()

        # is id defined?
        if item.get('id') is None:
            # label max size is 15, then 15 - 5 (size from string "vip_") = 10 symbols we can use
            source_string = item.get('interface', 'auto') + item['ip']
            label_size = cluster.globals['keepalived']['defaults']['label_size']
            item['id'] = hashlib.md5(source_string.encode('utf-8')).hexdigest()[:label_size]

        # is password defined?
        if item.get('password') is None:
            password_size = cluster.globals['keepalived']['defaults']['password_size']
            item['password'] = ("%032x" % random.getrandbits(128))[:password_size]

        # if nothing defined then use default names
        if item.get('hosts') is None:
            # Assign default list of all the balancer names. It can be an empty list.
            item['hosts'] = list(balancers)

        for j, record in enumerate(item['hosts']):
            if isinstance(record, str):
                item['hosts'][j] = record = {
                    'name': record
                }
            if record['name'] not in balancers:
                cluster.log.warning(f"Host {record['name']!r} for VRRP IP {item['ip']} is not among the balancers. "
                                    f"This VRRP IP will not be installed on this host.")
                continue
            if not record.get('priority'):
                priority_settings = static.GLOBALS['keepalived']['defaults']['priority']
                record['priority'] = priority_settings['max_value'] - (j + priority_settings['step'])
            if not record.get('interface') and item.get('interface'):
                record['interface'] = item['interface']
            if record.get('interface', 'auto') == 'auto':
                record['interface'] = autodetect_interface(cluster, record['name'])


def get_all_balancer_names(inventory: dict) -> List[str]:
    default_names = []

    # well, vrrp_ips is not empty, let's find balancers defined in config-file
    for node in inventory['nodes']:
        if 'balancer' in node['roles']:
            default_names.append(node['name'])

    return default_names


@enrichment(EnrichmentStage.FULL)
def enrich_inventory_calculate_nodegroup(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory

    # if vrrp_ips is empty, then nothing to do
    if not inventory['vrrp_ips']:
        return

    # Calculate group, where keepalived should be installed:
    names = []

    for item in inventory['vrrp_ips']:
        for record in item['hosts']:
            names.append(record['name'])

    # Create new group from balancers with Keepalived (to be) on them.
    keepalived_group = cluster.make_group_from_roles(['balancer']).new_group(apply_filter={
        'name': names
    })

    if not keepalived_group.is_empty():
        cluster.nodes['keepalived'] = keepalived_group


def install(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    defer = group.new_defer()
    collector = CollectorCallback(cluster)
    for node in defer.get_ordered_members_list():
        executable_name = cluster.get_package_association_for_node(
            node.get_host(), 'keepalived', 'executable_name')
        node.sudo("%s -v" % executable_name, warn=True, callback=collector)

    defer.flush()

    if not collector.result.is_any_failed():
        log.debug("Keepalived already installed, nothing to install")
    else:
        collector = CollectorCallback(cluster)
        for node in defer.get_ordered_members_list():
            package_name = cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'package_name')
            packages.install(node, include=package_name, pty=True, callback=collector)

        defer.flush()

    for node in defer.get_ordered_members_list():
        service_name = cluster.get_package_association_for_node(
            node.get_host(), 'keepalived', 'service_name')
        patch_path = "./resources/drop_ins/keepalived.conf"
        node.call(system.patch_systemd_service, service_name=service_name, patch_source=patch_path)
        node.call(install_haproxy_check_script)
        enable(node)

    defer.flush()
    return collector.result


def install_haproxy_check_script(group: DeferredGroup) -> None:
    script = utils.read_internal("./resources/scripts/check_haproxy.sh")
    group.put(io.StringIO(script), "/usr/local/bin/check_haproxy.sh", sudo=True)
    group.sudo("chmod +x /usr/local/bin/check_haproxy.sh")


def restart(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    cluster.log.debug("Restarting keepalived in all group...")
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'service_name')
            system.restart_service(node, name=service_name)

    cluster.log.debug("Sleep while keepalived comes-up...")
    time.sleep(static.GLOBALS['keepalived']['restart_wait'])


def enable(node: DeferredGroup) -> None:
    # currently it is invoked only for single node
    service_name = node.cluster.get_package_association_for_node(
        node.get_host(), 'keepalived', 'service_name')
    system.enable_service(node, name=service_name, now=True)


def disable(group: NodeGroup) -> None:
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = exe.cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'service_name')
            system.disable_service(node, name=service_name)


def generate_config(cluster: KubernetesCluster, node: NodeConfig) -> str:
    config_options: dict = cluster.inventory['services']['loadbalancer']['keepalived']
    config_string: Optional[str] = config_options.get('config')
    if config_string is not None:
        return config_string

    vrrps_ips = []
    keepalived_nodes = cluster.nodes['keepalived'].get_ordered_members_configs_list()
    for item in cluster.inventory['vrrp_ips']:
        host = next((record for record in item['hosts'] if record['name'] == node['name']), None)
        if not host:
            # This VRRP IP should not be configured on this node.
            # There is still at least one VRRP IP to configure on this node
            # due to the way how 'keepalived' group is calculated.
            continue
        vrrps_ips.append({
            'id': item['id'],
            'router_id': item['router_id'],
            'ip': item['ip'],
            'password': item['password'],
            'interface': host['interface'],
            'priority': host['priority'],
            'source': node['internal_address'],
            'peers': [
                i_node['internal_address'] for i_node in keepalived_nodes
                if any(i_node['name'] == record['name'] and i_node['internal_address'] != node['internal_address']
                       for record in item['hosts'])
            ]
        })

    if config_options.get('config_file'):
        config_source = utils.read_external(config_options['config_file'])
    else:
        config_source = utils.read_internal('templates/keepalived.conf.j2')
    config = Template(config_source).render(vrrp_ips=vrrps_ips, globals=config_options['global'])

    return config


def configure(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            node_name = node.get_node_name()
            log.debug("Configuring keepalived on '%s'..." % node_name)

            config_location = cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'config_location')

            config = generate_config(cluster, node.get_config())
            utils.dump_file(cluster, config, 'keepalived_%s.conf' % node_name)

            node.put(io.StringIO(config), config_location, sudo=True, mkdir=True)
            node.sudo('ls -la %s' % config_location, callback=collector)

    log.debug(collector.result)

    restart(group)

    return status(group)


def status(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = cluster.get_package_association_for_node(
                node.get_host(), 'keepalived', 'service_name')
            system.service_status(node, name=service_name, callback=collector)

    return collector.result



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC, abstractmethod

from kubemarine.core import log
from kubemarine.core.connections import ConnectionPool


class Environment(ABC):
    @property
    @abstractmethod
    def connection_pool(self) -> ConnectionPool:
        pass

    @property
    @abstractmethod
    def log(self) -> log.EnhancedLogger:
        pass

    @staticmethod
    def is_deploying_from_windows() -> bool:
        return os.name == 'nt'



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent
from typing import List, Optional, Dict

import yaml

from kubemarine.core import log
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest, Identity


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    if not inventory["plugins"]["local-path-provisioner"]["install"]:
        return

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_plugin = cluster.raw_inventory.get("plugins", {}).get("local-path-provisioner", {})
    if "resources" in raw_plugin:
        inventory["plugins"]["local-path-provisioner"]["resources"] = raw_plugin["resources"]


class LocalPathProvisionerManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None) -> None:
        super().__init__(logger, inventory, Identity('local-path-provisioner'), original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_local-path-storage",
            "ServiceAccount_local-path-provisioner-service-account",
            "Role_local-path-provisioner-role",
            "ClusterRole_local-path-provisioner-role",
            "RoleBinding_local-path-provisioner-bind",
            "ClusterRoleBinding_local-path-provisioner-bind",
            "Deployment_local-path-provisioner",
            "StorageClass_local-path",
            "ConfigMap_local-path-config",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_local_path_storage,
            self.enrich_service_account,
            self.enrich_service_account_secret,
            self.enrich_deployment_local_path_provisioner,
            self.enrich_storageclass_local_path,
            self.enrich_configmap_local_path_config,
        ]

    def get_namespace_to_necessary_pss_profiles(self) -> Dict[str, str]:
        return {'local-path-storage': 'privileged'}

    def enrich_namespace_local_path_storage(self, manifest: Manifest) -> None:
        self.assign_default_pss_labels(manifest, 'local-path-storage')

    def enrich_service_account_secret(self, manifest: Manifest) -> None:
        new_yaml = yaml.safe_load(service_account_secret)

        service_account_key = "ServiceAccount_local-path-provisioner-service-account"
        service_account_index = manifest.all_obj_keys().index(service_account_key) \
            if service_account_key in manifest.all_obj_keys() else -1
        
        self.include(manifest, service_account_index + 1, new_yaml)

    def enrich_service_account(self, manifest: Manifest) -> None:
        key = "ServiceAccount_local-path-provisioner-service-account"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['automountServiceAccountToken'] = False

    def enrich_deployment_local_path_provisioner(self, manifest: Manifest) -> None:
        service_account_name = "local-path-provisioner-service-account"
        key = "Deployment_local-path-provisioner"

        source_yaml = manifest.get_obj(key, patch=True)
        self.enrich_volume_and_volumemount(source_yaml, service_account_name)
        self.log.verbose(f"The {key} has been updated to include the new secret volume and mount.")

        self.enrich_image_for_container(manifest, key,
            container_name='local-path-provisioner', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='local-path-provisioner')
        self.enrich_tolerations(manifest, key)

    def enrich_storageclass_local_path(self, manifest: Manifest) -> None:
        key = "StorageClass_local-path"
        source_yaml = manifest.get_obj(key, patch=True)
        metadata = source_yaml['metadata']

        is_default = str(self.inventory['plugins']['local-path-provisioner']['storage-class']['is-default'])
        metadata.setdefault('annotations', {})['storageclass.kubernetes.io/is-default-class'] = is_default
        self.log.verbose(f"The {key} has been patched in 'metadata.annotations' "
                         f"with 'storageclass.kubernetes.io/is-default-class: {is_default}'")

        name = self.inventory['plugins']['local-path-provisioner']['storage-class']['name']
        metadata['name'] = self.inventory['plugins']['local-path-provisioner']['storage-class']['name']
        self.log.verbose(f"The {key} has been patched in 'metadata.name' with {name!r}")

    def enrich_configmap_local_path_config(self, manifest: Manifest) -> None:
        key = "ConfigMap_local-path-config"
        source_yaml = manifest.get_obj(key, patch=True)
        data = source_yaml['data']

        config_json = data['config.json']
        config_json = config_json.replace('/opt/local-path-provisioner',
                                          self.inventory['plugins']['local-path-provisioner']['volume-dir'])
        data['config.json'] = config_json
        config_json_oneline = config_json.replace('\n', ' ')
        self.log.verbose(f"The {key} has been patched in 'data.config.json' with {config_json_oneline!r}")

        helperpod_yaml_str = data['helperPod.yaml']
        helperpod_yaml = yaml.safe_load(helperpod_yaml_str)
        busybox_source_image = helperpod_yaml['spec']['containers'][0]['image']
        helperpod_yaml_str = helperpod_yaml_str.replace(
            busybox_source_image, self.get_target_image(image_key='helper-pod-image'))
        data['helperPod.yaml'] = helperpod_yaml_str
        helperpod_yaml_str_oneline = helperpod_yaml_str.replace('\n', ' ')
        self.log.verbose(f"The {key} has been patched in 'data.helperPod.yaml' with {helperpod_yaml_str_oneline!r}")


service_account_secret = dedent("""\
    apiVersion: v1
    kind: Secret
    metadata:
      name: local-path-provisioner-service-account-token
      namespace: local-path-storage
      annotations:
        kubernetes.io/service-account.name: local-path-provisioner-service-account
    type: kubernetes.io/service-account-token  
""")



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict

from kubemarine.core import utils
from kubemarine.core.cluster import EnrichmentStage, enrichment, KubernetesCluster
from kubemarine.core.yaml_merger import default_merger


@enrichment(EnrichmentStage.PROCEDURE, procedures=['reconfigure'])
def enrich_reconfigure_inventory(cluster: KubernetesCluster) -> None:
    # Do not apply usual merging strategy, always merge and never override.
    # This is because `reconfigure` might support to change only subset of sections allowed during installation.
    # Append reconfiguring patches to the end for them to have priority.
    procedure_patches = cluster.procedure_inventory.get('patches', [])
    if procedure_patches:
        cluster.inventory.setdefault('patches', []).extend(procedure_patches)


@enrichment(EnrichmentStage.FULL)
def verify_inventory_patches(cluster: KubernetesCluster) -> None:
    for i, patch in enumerate(cluster.inventory['patches']):
        if patch.get('nodes') is not None:
            all_nodes_names = cluster.nodes['all'].get_nodes_names()
            unknown_nodes = set(patch['nodes']) - set(all_nodes_names)
            if unknown_nodes:
                # Only warn instead of raising an error to allow remove & add the same node.
                cluster.log.warning(
                    f"Unknown node names {', '.join(map(repr, unknown_nodes))} "
                    f"provided for inventory{utils.pretty_path(['patches', i])}. ")


def patch_inventory(cluster: KubernetesCluster, section_names: List[str]) -> None:
    """
    Take nested `section_names` from each inventory patch,
    and merge with nested `section_names` of the main inventory.

    Fill `KubernetesCluster.node_inventory` with the result.

    :param cluster: KubernetesCluster instance
    :param section_names: nested sections of the inventory
    """
    same_configs: Dict[tuple, dict] = {}
    for node in cluster.nodes['all'].get_ordered_members_list():
        patch_ids = []
        for i, patch in enumerate(cluster.inventory['patches']):
            group = cluster.create_group_from_groups_nodes_names(
                patch.get('groups', []), patch.get('nodes', []))

            if group.has_node(node.get_node_name()):
                patch_ids.append(i)

        same_id = tuple(patch_ids)
        if same_id not in same_configs:
            config = utils.deepcopy_yaml(_extract_config(cluster.inventory, section_names))

            for i in same_id:
                nxt = utils.deepcopy_yaml(_extract_config(cluster.inventory['patches'][i], section_names))
                config = default_merger.merge(config, nxt)

            same_configs[same_id] = config

        node_inventory = cluster.nodes_inventory.setdefault(node.get_host(), {})
        config = same_configs[same_id]

        _set_config(node_inventory, section_names, config)


def _extract_config(inventory: dict, section_names: List[str]) -> dict:
    section = inventory
    for name in section_names:
        section = section.get(name, {})

    return section


def _set_config(inventory: dict, section_names: List[str], config: dict) -> None:
    section = inventory
    for name in section_names[:-1]:
        section = section.setdefault(name, {})

    section[section_names[-1]] = config



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import re
from textwrap import dedent
from typing import List, Optional, Dict, Callable, Sequence, Union

import yaml
from jinja2 import Template
from ordered_set import OrderedSet

from kubemarine import plugins, system
from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, DeferredGroup, CollectorCallback, AbstractGroup, RunResult
from kubemarine.core.yaml_merger import override_merger
from kubemarine.kubernetes.object import KubernetesObject

ERROR_WAIT_FOR_PODS_NOT_SUPPORTED = "Waiting for pods of {components} components is currently not supported"
ERROR_RESTART_NOT_SUPPORTED = "Restart of {components} components is currently not supported"
ERROR_RECONFIGURE_NOT_SUPPORTED = "Reconfiguration of {components} components is currently not supported"

COMPONENTS_CONSTANTS: dict = {
    'kube-apiserver/cert-sans': {
        'sections': [
            ['services', 'kubeadm', 'apiServer', 'certSANs'],
        ],
        'init_phase': 'certs apiserver',
    },
    'kube-apiserver': {
        'sections': [
            ['services', 'kubeadm', 'apiServer'],
            ['services', 'kubeadm_patches', 'apiServer'],
        ],
        'patch': {
            'section': 'apiServer',
            'target_template': r'kube-apiserver\*',
            'file': 'kube-apiserver+json.json'
        },
        'init_phase': 'control-plane apiserver',
    },
    'kube-scheduler': {
        'sections': [
            ['services', 'kubeadm', 'scheduler'],
            ['services', 'kubeadm_patches', 'scheduler'],
        ],
        'patch': {
            'section': 'scheduler',
            'target_template': r'kube-scheduler\*',
            'file': 'kube-scheduler+json.json'
        },
        'init_phase': 'control-plane scheduler',
    },
    'kube-controller-manager': {
        'sections': [
            ['services', 'kubeadm', 'controllerManager'],
            ['services', 'kubeadm_patches', 'controllerManager'],
        ],
        'patch': {
            'section': 'controllerManager',
            'target_template': r'kube-controller-manager\*',
            'file': 'kube-controller-manager+json.json'
        },
        'init_phase': 'control-plane controller-manager',
    },
    'etcd': {
        'sections': [
            ['services', 'kubeadm', 'etcd'],
            ['services', 'kubeadm_patches', 'etcd'],
        ],
        'patch': {
            'section': 'etcd',
            'target_template': r'etcd\*',
            'file': 'etcd+json.json'
        },
        'init_phase': 'etcd local',
    },
    'kubelet': {
        'sections': [
            ['services', 'kubeadm_kubelet'],
            ['services', 'kubeadm_patches', 'kubelet'],
        ],
        'patch': {
            'section': 'kubelet',
            'target_template': r'kubeletconfiguration\*',
            'file': 'kubeletconfiguration.yaml'
        }
    },
    'kube-proxy': {
        'sections': [
            ['services', 'kubeadm_kube-proxy'],
        ]
    },
}

CONFIGMAPS_CONSTANTS = {
    'kubeadm-config': {
        'section': 'kubeadm',
        'key': 'ClusterConfiguration',
        'init_phase': 'upload-config kubeadm',
    },
    'kubelet-config': {
        'section': 'kubeadm_kubelet',
        'key': 'kubelet',
        'init_phase': 'upload-config kubelet',
    },
    'kube-proxy': {
        'section': 'kubeadm_kube-proxy',
        'key': 'config.conf',
    },
}


CONTROL_PLANE_COMPONENTS = ["kube-apiserver", "kube-scheduler", "kube-controller-manager", "etcd"]
CONTROL_PLANE_SPECIFIC_COMPONENTS = ['kube-apiserver/cert-sans'] + CONTROL_PLANE_COMPONENTS
NODE_COMPONENTS = ["kubelet", "kube-proxy"]
ALL_COMPONENTS = CONTROL_PLANE_SPECIFIC_COMPONENTS + NODE_COMPONENTS
COMPONENTS_SUPPORT_PATCHES = CONTROL_PLANE_COMPONENTS + ['kubelet']


class KubeadmConfig:
    def __init__(self, cluster: KubernetesCluster):
        self.cluster = cluster

        inventory = cluster.inventory
        self.maps: Dict[str, dict] = {
            configmap: inventory["services"][constants['section']]
            for configmap, constants in CONFIGMAPS_CONSTANTS.items()}

        self.loaded_maps: Dict[str, KubernetesObject] = {}

    def is_loaded(self, configmap: str) -> bool:
        return configmap in self.loaded_maps

    def load(self, configmap: str, control_plane: NodeGroup, edit_func: Callable[[dict], dict] = None) -> dict:
        """
        Load ConfigMap as object, retrieve plain config from it, and apply `edit_func` to the plain config if provided.

        :param configmap: name of ConfigMap
        :param edit_func: function to apply changes
        :param control_plane: Use this control plane node to fetch the ConfigMap.
        """
        configmap_obj = KubernetesObject(self.cluster, 'ConfigMap', configmap, 'kube-system')
        configmap_obj.reload(control_plane)

        self.loaded_maps[configmap] = configmap_obj

        key = CONFIGMAPS_CONSTANTS[configmap]['key']
        config: dict = yaml.safe_load(configmap_obj.obj["data"][key])

        if edit_func is not None:
            config = edit_func(config)
            configmap_obj.obj["data"][key] = yaml.dump(config)

        self.maps[configmap] = config
        return config

    def apply(self, configmap: str, control_plane: NodeGroup) -> None:
        """
        Apply ConfigMap that was previously changed using edit().

        :param configmap: name of ConfigMap
        :param control_plane: Use this control plane node to apply the ConfigMap.
        :return:
        """
        if not self.is_loaded(configmap):
            raise ValueError(f"To apply changed {configmap} ConfigMap, it is necessary to fetch it first")

        self.loaded_maps[configmap].apply(control_plane)

    def to_yaml(self, init_config: dict) -> str:
        configs = list(self.maps.values())
        configs.append(init_config)
        return yaml.dump_all(configs)

    def merge_with_inventory(self, configmap: str) -> Callable[[dict], dict]:
        def merge_func(config_: dict) -> dict:
            patch_config: dict = KubeadmConfig(self.cluster).maps[configmap]
            # It seems that all default lists are always overridden with custom instead of appending,
            # and so override merger seems the most suitable.
            config_ = override_merger.merge(config_, utils.deepcopy_yaml(patch_config))
            return config_

        return merge_func


def kubeadm_extended_dryrun(cluster: KubernetesCluster) -> bool:
    return kubernetes_minor_release_at_least(cluster.inventory, "v1.26")


def is_container_runtime_not_configurable(cluster: KubernetesCluster) -> bool:
    return kubernetes_minor_release_at_least(cluster.inventory, "v1.27")


def kube_proxy_overwrites_higher_system_values(cluster: KubernetesCluster) -> bool:
    return kubernetes_minor_release_at_least(cluster.inventory, "v1.29")


def kubernetes_minor_release_at_least(inventory: dict, minor_version: str) -> bool:
    kubernetes_version = inventory["services"]["kubeadm"]["kubernetesVersion"]
    return utils.version_key(kubernetes_version)[0:2] >= utils.minor_version_key(minor_version)


def get_init_config(cluster: KubernetesCluster, group: AbstractGroup[RunResult], *,
                    init: bool, join_dict: dict = None) -> dict:
    inventory = cluster.inventory

    if join_dict is None:
        join_dict = {}

    init_kind = 'InitConfiguration' if init else 'JoinConfiguration'

    control_plane_spec = {}
    if group.nodes_amount() > 1:
        control_planes = group.new_group(lambda node: 'control-plane' in node['roles'])
        if not control_planes.is_empty():
            raise Exception("Init/Join configuration for control planes should be unique")

        control_plane = False
        worker = True
    else:
        node_config = group.get_config()
        control_plane = 'control-plane' in node_config['roles']
        worker = 'worker' in node_config['roles']
        if control_plane:
            control_plane_spec = {'localAPIEndpoint': {
                'advertiseAddress': node_config['internal_address']
            }}

    init_config: dict = {
        'apiVersion': inventory["services"]["kubeadm"]['apiVersion'],
        'kind': init_kind,
        'patches': {'directory': '/etc/kubernetes/patches'},
    }
    if init:
        if control_plane:
            init_config.update(control_plane_spec)
    else:
        if control_plane:
            control_plane_spec['certificateKey'] = join_dict['certificate-key']
            init_config['controlPlane'] = control_plane_spec

        init_config['discovery'] = {
            'bootstrapToken': {
                'apiServerEndpoint': inventory["services"]["kubeadm"]['controlPlaneEndpoint'],
                'token': join_dict['token'],
                'caCertHashes': [
                    join_dict['discovery-token-ca-cert-hash']
                ]
            }
        }

    if control_plane and worker:
        init_config.setdefault('nodeRegistration', {})['taints'] = []

    _configure_container_runtime(cluster, init_config)

    return init_config


def get_kubeadm_config(cluster: KubernetesCluster, init_config: dict) -> str:
    return KubeadmConfig(cluster).to_yaml(init_config)


def _configure_container_runtime(cluster: KubernetesCluster, kubeadm_config: dict) -> None:
    kubelet_extra_args = kubeadm_config.setdefault('nodeRegistration', {}).setdefault('kubeletExtraArgs', {})

    kubeadm_config['nodeRegistration']['criSocket'] = '/var/run/containerd/containerd.sock'

    if not is_container_runtime_not_configurable(cluster):
        kubelet_extra_args['container-runtime'] = 'remote'

    kubelet_extra_args['container-runtime-endpoint'] = 'unix:///run/containerd/containerd.sock'


def reconfigure_components(group: NodeGroup, components: List[str],
                           *,
                           edit_functions: Dict[str, Callable[[dict], dict]] = None,
                           force_restart: bool = False) -> None:
    """
    Reconfigure the specified `components` on `group` of nodes.
    Control-plane nodes are reconfigured first.
    The cluster is not required to be working to update control plane manifests.

    :param group: nodes to reconfigure components on
    :param components: List of control plane components or `kube-proxy`, or `kubelet` to reconfigure.
    :param edit_functions: Callables that edit the specified kubeadm-managed ConfigMaps in a custom way.
                           The ConfigMaps are fetched from the first control plane,
                           instead of being generated.
                           This implies necessity of working API server.
    :param force_restart: Restart the given `components` even if nothing has changed in their configuration.
    """
    not_supported = list(OrderedSet[str](components) - set(ALL_COMPONENTS))
    if not_supported:
        raise Exception(ERROR_RECONFIGURE_NOT_SUPPORTED.format(components=not_supported))

    if edit_functions is None:
        edit_functions = {}

    cluster: KubernetesCluster = group.cluster
    logger = cluster.log

    control_planes = (cluster.nodes['control-plane']
                      .intersection_group(group))
    workers = (cluster.make_group_from_roles(['worker'])
               .exclude_group(control_planes)
               .intersection_group(group))

    group = control_planes.include_group(workers)
    if group.is_empty():
        logger.debug("No Kubernetes nodes to reconfigure components")
        return

    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    if not control_planes.is_empty():
        first_control_plane = control_planes.get_first_member()

    timestamp = utils.get_current_timestamp_formatted()
    backup_dir = '/etc/kubernetes/tmp/kubemarine-backup-' + timestamp
    logger.debug(f"Using backup directory {backup_dir}")

    kubeadm_config = KubeadmConfig(cluster)
    for configmap, func in edit_functions.items():
        kubeadm_config.load(configmap, first_control_plane, func)

    # This configuration will be used for `kubeadm init phase upload-config` phase.
    # InitConfiguration is necessary to specify patches directory.
    # Use patches from the first control plane, to make it the same as during regular installation.
    # Patches are used to upload KubeletConfiguration for some reason (which is likely a gap)
    # https://github.com/kubernetes/kubernetes/issues/123090
    upload_config = '/etc/kubernetes/upload-config.yaml'
    upload_config_uploaded = False

    def prepare_upload_config() -> None:
        nonlocal upload_config_uploaded
        if upload_config_uploaded:
            return

        logger.debug(f"Uploading cluster config to control plane: {first_control_plane.get_node_name()}")
        _upload_config(cluster, first_control_plane, kubeadm_config, upload_config)

        upload_config_uploaded = True

    # This configuration will be generated for control plane nodes,
    # and will be used to `kubeadm init phase control-plane / etcd / certs`.
    reconfigure_config = '/etc/kubernetes/reconfigure-config.yaml'

    _prepare_nodes_to_reconfigure_components(cluster, group, components,
                                             kubeadm_config, reconfigure_config, backup_dir)

    kubeadm_config_updated = False
    kubelet_config_updated = False
    kube_proxy_config_updated = False
    kube_proxy_changed = False
    for node in (control_planes.get_ordered_members_list() + workers.get_ordered_members_list()):
        _components = _choose_components(node, components)
        if not _components:
            continue

        # Firstly reconfigure components that do not require working API server
        control_plane_components = list(OrderedSet[str](_components) & set(CONTROL_PLANE_SPECIFIC_COMPONENTS))
        if control_plane_components:
            _reconfigure_control_plane_components(cluster, node, control_plane_components, force_restart,
                                                  reconfigure_config, backup_dir)

            # Upload kubeadm-config after control plane components are successfully reconfigured on the first node.
            if not kubeadm_config_updated:
                prepare_upload_config()
                _update_configmap(cluster, first_control_plane, 'kubeadm-config',
                                  _configmap_init_phase_uploader('kubeadm-config', upload_config),
                                  backup_dir)
                kubeadm_config_updated = True

        node_components = list(OrderedSet[str](_components) & set(NODE_COMPONENTS))
        if node_components:
            if 'kubelet' in node_components and not kubelet_config_updated:
                prepare_upload_config()
                _update_configmap(cluster, first_control_plane, 'kubelet-config',
                                  _configmap_init_phase_uploader('kubelet-config', upload_config),
                                  backup_dir)

                kubelet_config_updated = True

            if 'kube-proxy' in node_components and not kube_proxy_config_updated:
                kube_proxy_changed = _update_configmap(cluster, first_control_plane, 'kube-proxy',
                                                       _kube_proxy_configmap_uploader(cluster, kubeadm_config),
                                                       backup_dir)

                kube_proxy_config_updated = True

            _reconfigure_node_components(cluster, node, node_components, force_restart, first_control_plane,
                                         kube_proxy_changed, backup_dir)


def restart_components(group: NodeGroup, components: List[str]) -> None:
    """
    Currently it is supported to restart only
    'kube-apiserver', 'kube-scheduler', 'kube-controller-manager', 'etcd'.

    :param group: nodes to restart components on
    :param components: Kubernetes components to restart
    """
    not_supported = list(OrderedSet[str](components) - set(CONTROL_PLANE_COMPONENTS))
    if not_supported:
        raise Exception(ERROR_RESTART_NOT_SUPPORTED.format(components=not_supported))

    cluster: KubernetesCluster = group.cluster

    for node in group.get_ordered_members_list():
        _components = _choose_components(node, components)
        _restart_containers(cluster, node, _components)
        wait_for_pods(node, _components)


def wait_for_pods(group: NodeGroup, components: Sequence[str] = None) -> None:
    """
    Wait for pods of Kubernetes components on the given `group` of nodes.
    All relevant components are waited for unless specific list of `components` is given.
    For nodes that are not control planes, only 'kube-proxy' can be waited for.

    :param group: nodes to wait pods on
    :param components: Kubernetes components to wait for.
    """
    if components is not None:
        if not components:
            return

        not_supported = list(OrderedSet[str](components) - set(CONTROL_PLANE_COMPONENTS) - {'kube-proxy'})
        if not_supported:
            raise Exception(ERROR_WAIT_FOR_PODS_NOT_SUPPORTED.format(components=not_supported))

    cluster: KubernetesCluster = group.cluster
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    expect_config = cluster.inventory['globals']['expect']['pods']['kubernetes']

    for node in group.get_ordered_members_list():
        node_name = node.get_node_name()
        is_control_plane = 'control-plane' in node.get_config()['roles']
        cluster.log.debug(f"Waiting for system pods on node: {node_name}")

        if components is not None:
            _components = list(components)
        else:
            _components = ['kube-proxy']
            if is_control_plane:
                _components.extend(CONTROL_PLANE_COMPONENTS)

        _components = _choose_components(node, _components)
        if not _components:
            continue

        control_plane = node
        if not is_control_plane:
            control_plane = first_control_plane

        plugins.expect_pods(cluster, _components, namespace='kube-system',
                            control_plane=control_plane, node_name=node_name,
                            timeout=expect_config['timeout'],
                            retries=expect_config['retries'])


# function to create kubeadm patches and put them to a node
def create_kubeadm_patches_for_node(cluster: KubernetesCluster, node: NodeGroup) -> None:
    cluster.log.verbose(f"Create and upload kubeadm patches to %s..." % node.get_node_name())
    node.sudo("mkdir -p /etc/kubernetes/patches")
    defer = node.new_defer()
    for component in COMPONENTS_SUPPORT_PATCHES:
        _create_kubeadm_patches_for_component_on_node(cluster, defer, component)

    defer.flush()


def patch_kubelet_configmap(control_plane: AbstractGroup[RunResult]) -> None:
    """
    Apply W/A until https://github.com/kubernetes/kubeadm/issues/3034 is resolved
    for all the supported Kubernetes versions.

    :param control_plane: control plane to operate on
    """
    # Make sure to check kubeadm_kubelet in the inventory.
    cluster: KubernetesCluster = control_plane.cluster
    kubeadm_config = KubeadmConfig(cluster)
    if 'resolvConf' in kubeadm_config.maps['kubelet-config']:
        return

    control_plane.sudo(
        f'kubectl get cm -n kube-system kubelet-config -o yaml '
        '| grep -v "resolvConf:" '
        '| sudo kubectl apply -f -')


def _upload_config(cluster: KubernetesCluster, control_plane: AbstractGroup[RunResult],
                   kubeadm_config: KubeadmConfig, remote_path: str,
                   *, patches_dir: str = '/etc/kubernetes/patches') -> None:
    name = remote_path.rstrip('.yaml').split('/')[-1]

    init_config = get_init_config(cluster, control_plane, init=True)
    init_config['patches']['directory'] = patches_dir
    config = kubeadm_config.to_yaml(init_config)
    utils.dump_file(cluster, config, f"{name}_{control_plane.get_node_name()}.yaml")

    control_plane.put(io.StringIO(config), remote_path, sudo=True)


def _update_configmap(cluster: KubernetesCluster, control_plane: NodeGroup, configmap: str,
                      uploader: Callable[[DeferredGroup], None], backup_dir: str) -> bool:
    logger = cluster.log

    logger.debug(f"Updating {configmap} ConfigMap")
    defer = control_plane.new_defer()
    collector = CollectorCallback(cluster)

    key = CONFIGMAPS_CONSTANTS[configmap]['key'].replace('.', r'\.')
    configmap_cmd = f'sudo kubectl get cm -n kube-system {configmap} -o=jsonpath="{{.data.{key}}}"'

    # backup
    defer.run(f'sudo mkdir -p {backup_dir}')
    backup_file = f'{backup_dir}/{configmap}.yaml'
    defer.run(f'(set -o pipefail && {configmap_cmd} | sudo tee {backup_file}) > /dev/null')

    # update
    uploader(defer)

    # compare
    defer.run(f'sudo cat {backup_file}', callback=collector)
    defer.run(configmap_cmd, callback=collector)

    defer.flush()

    results = collector.results[defer.get_host()]
    return _detect_changes(logger, results[0].stdout, results[1].stdout,
                           fromfile=backup_file, tofile=f'{configmap} ConfigMap')


def _configmap_init_phase_uploader(configmap: str, upload_config: str) -> Callable[[DeferredGroup], None]:
    def upload(control_plane: DeferredGroup) -> None:
        init_phase = CONFIGMAPS_CONSTANTS[configmap]['init_phase']
        control_plane.run(f'sudo kubeadm init phase {init_phase} --config {upload_config}')
        if configmap == 'kubelet-config':
            patch_kubelet_configmap(control_plane)

    return upload


def _kube_proxy_configmap_uploader(cluster: KubernetesCluster, kubeadm_config: KubeadmConfig) \
        -> Callable[[DeferredGroup], None]:

    # Unfortunately, there is no suitable kubeadm command to upload the generated ConfigMap.
    # This makes it impossible to reset some property to default by deleting it from `services.kubeadm_kube-proxy`.

    def upload(control_plane_deferred: DeferredGroup) -> None:
        control_plane_deferred.flush()
        control_plane = cluster.make_group(control_plane_deferred.get_hosts())

        # reconfigure_components() can be called with custom editing function.
        # The ConfigMap is already fetched and changed.
        if not kubeadm_config.is_loaded('kube-proxy'):
            kubeadm_config.load('kube-proxy', control_plane, kubeadm_config.merge_with_inventory('kube-proxy'))

        # Apply updated kube-proxy ConfigMap
        kubeadm_config.apply('kube-proxy', control_plane)

    return upload


def _choose_components(node: AbstractGroup[RunResult], components: List[str]) -> List[str]:
    roles = node.get_config()['roles']

    return [c for c in components if c in NODE_COMPONENTS and set(roles) & {'control-plane', 'worker'}
            or 'control-plane' in roles and c in CONTROL_PLANE_SPECIFIC_COMPONENTS]


def _prepare_nodes_to_reconfigure_components(cluster: KubernetesCluster, group: NodeGroup, components: List[str],
                                             kubeadm_config: KubeadmConfig,
                                             reconfigure_config: str, backup_dir: str) -> None:
    logger = cluster.log
    defer = group.new_defer()
    for node in defer.get_ordered_members_list():
        _components = _choose_components(node, components)
        if not _components:
            continue

        if set(_components) & set(CONTROL_PLANE_SPECIFIC_COMPONENTS):
            logger.debug(f"Uploading config for control plane components on node: {node.get_node_name()}")
            _upload_config(cluster, node, kubeadm_config, reconfigure_config)

        if set(_components) & set(COMPONENTS_SUPPORT_PATCHES):
            node.sudo("mkdir -p /etc/kubernetes/patches")
            node.sudo(f'mkdir -p {backup_dir}/patches')

        if set(_components) & set(CONTROL_PLANE_COMPONENTS):
            node.sudo(f'mkdir -p {backup_dir}/manifests')

        for component in _components:
            if component not in COMPONENTS_SUPPORT_PATCHES:
                continue

            _create_kubeadm_patches_for_component_on_node(cluster, node, component, backup_dir)

    defer.flush()


def _reconfigure_control_plane_components(cluster: KubernetesCluster, node: NodeGroup, components: List[str],
                                          force_restart: bool,
                                          reconfigure_config: str, backup_dir: str) -> None:
    logger = cluster.log
    logger.debug(f"Reconfiguring control plane components {components} on node: {node.get_node_name()}")

    defer = node.new_defer()

    containers_restart = OrderedSet[str]()
    pods_wait = OrderedSet[str]()
    for component in components:
        if component == 'kube-apiserver/cert-sans':
            _reconfigure_apiserver_certsans(defer, reconfigure_config, backup_dir)
            containers_restart.add('kube-apiserver')
            pods_wait.add('kube-apiserver')
            continue

        # Let's anyway wait for pods as the component may be broken due to previous runs.
        pods_wait.add(component)
        if (_reconfigure_control_plane_component(cluster, defer, component, reconfigure_config, backup_dir)
                or force_restart):
            containers_restart.add(component)
        else:
            # Manifest file may be changed in formatting but not in meaningful content.
            # Kubelet is not observed to restart the container in this case, neither will we.
            pass

    defer.flush()
    _restart_containers(cluster, node, containers_restart)
    wait_for_pods(node, pods_wait)


def _reconfigure_apiserver_certsans(node: DeferredGroup, reconfigure_config: str, backup_dir: str) -> None:
    apiserver_certs = r'find /etc/kubernetes/pki/ -name apiserver.\*'

    # backup
    node.sudo(f'mkdir -p {backup_dir}/pki')
    node.sudo(f'{apiserver_certs} -exec cp {{}} {backup_dir}/pki \\;')

    # create cert
    node.sudo(f'{apiserver_certs} -delete')
    init_phase = COMPONENTS_CONSTANTS['kube-apiserver/cert-sans']['init_phase']
    node.sudo(f'kubeadm init phase {init_phase} --config {reconfigure_config}', pty=True)


def _reconfigure_control_plane_component(cluster: KubernetesCluster, node: DeferredGroup, component: str,
                                         reconfigure_config: str, backup_dir: str) -> bool:
    manifest = f'/etc/kubernetes/manifests/{component}.yaml'
    backup_file = f'{backup_dir}/manifests/{component}.yaml'
    collector = CollectorCallback(cluster)

    # backup
    node.sudo(f"cp {manifest} {backup_file}")

    # update
    init_phase = COMPONENTS_CONSTANTS[component]['init_phase']
    node.sudo(f'kubeadm init phase {init_phase} --config {reconfigure_config}', pty=True)

    # compare
    node.sudo(f'cat {backup_file}', callback=collector)
    node.sudo(f'cat {manifest}', callback=collector)

    node.flush()

    results = collector.results[node.get_host()]
    return _detect_changes(cluster.log, results[0].stdout, results[1].stdout,
                           fromfile=backup_file, tofile=manifest)


def _reconfigure_node_components(cluster: KubernetesCluster, node: NodeGroup, components: List[str],
                                 force_restart: bool, control_plane: NodeGroup,
                                 kube_proxy_changed: bool, backup_dir: str) -> None:
    logger = cluster.log

    is_control_plane = 'control-plane' in node.get_config()['roles']
    kube_proxy_restart = kube_proxy_changed or force_restart
    containers_restart = OrderedSet[str]()
    pods_wait = OrderedSet[str]()
    if 'kube-proxy' in components:
        pods_wait.add('kube-proxy')

    if 'kubelet' in components:
        logger.debug(f"Reconfiguring kubelet on node: {node.get_node_name()}")

        pods_wait.add('kube-proxy')
        if is_control_plane:
            pods_wait.update(CONTROL_PLANE_COMPONENTS)

        defer = node.new_defer()
        if _reconfigure_kubelet(cluster, defer, backup_dir) or force_restart:
            system.restart_service(defer, 'kubelet')
            # It is not clear how to check that kubelet is healthy.
            # Let's restart and check health of all components.
            kube_proxy_restart = True
            if is_control_plane:
                # No need to manually kill container for kube-proxy. It is restarted as soon as pod is deleted.
                containers_restart.update(CONTROL_PLANE_COMPONENTS)

        defer.flush()

    # Delete pod for 'kube-proxy' early while 'kube-apiserver' is expected to be available.
    if kube_proxy_restart:
        _delete_pods(cluster, node, control_plane, ['kube-proxy'])

    _restart_containers(cluster, node, containers_restart)
    wait_for_pods(node, pods_wait)


def _reconfigure_kubelet(cluster: KubernetesCluster, node: DeferredGroup,
                         backup_dir: str) -> bool:
    config = '/var/lib/kubelet/config.yaml'
    backup_file = f'{backup_dir}/kubelet/config.yaml'
    collector = CollectorCallback(cluster)

    # backup
    node.sudo(f'mkdir -p {backup_dir}/kubelet')
    node.sudo(f"cp {config} {backup_file}")

    # update
    node.sudo(f'kubeadm upgrade node phase kubelet-config --patches=/etc/kubernetes/patches', pty=True)

    # compare
    node.sudo(f'cat {backup_file}', callback=collector)
    node.sudo(f'cat {config}', callback=collector)

    node.flush()

    results = collector.results[node.get_host()]
    return _detect_changes(cluster.log, results[0].stdout, results[1].stdout,
                           fromfile=backup_file, tofile=config)


def compare_manifests(cluster: KubernetesCluster, *, with_inventory: bool) \
        -> Dict[str, Dict[str, Optional[str]]]:
    """
    Generate manifests in dry-run mode for all control plane components on all control plane nodes,
    and compare with already present manifests.

    :param cluster: KubernetesCluster instance
    :param with_inventory: flag if cluster configuration should be generated from the inventory
    :return: mapping host -> component -> diff string
    """
    kubeadm_config = KubeadmConfig(cluster)
    if not with_inventory:
        kubeadm_config.load('kubeadm-config', cluster.nodes['control-plane'].get_first_member())

    control_planes = cluster.nodes['control-plane'].new_defer()
    temp_config = utils.get_remote_tmp_path()
    patches_dir = '/etc/kubernetes/patches'
    if with_inventory:
        patches_dir = utils.get_remote_tmp_path()

    components = [c for c in CONTROL_PLANE_COMPONENTS
                  if c != 'etcd' or kubeadm_extended_dryrun(cluster)]

    tmp_dirs_cmd = "sh -c 'sudo ls /etc/kubernetes/tmp/ | grep dryrun 2>/dev/null || true'"
    old_tmp_dirs = CollectorCallback(cluster)
    new_tmp_dirs = CollectorCallback(cluster)
    for defer in control_planes.get_ordered_members_list():
        _upload_config(cluster, defer, kubeadm_config, temp_config, patches_dir=patches_dir)
        if with_inventory:
            defer.sudo(f'mkdir -p {patches_dir}')

        for component in components:
            if with_inventory:
                _create_kubeadm_patches_for_component_on_node(cluster, defer, component,
                                                              patches_dir=patches_dir, reset=False)

            defer.sudo(tmp_dirs_cmd, callback=old_tmp_dirs)

            init_phase = COMPONENTS_CONSTANTS[component]['init_phase']
            defer.sudo(f'kubeadm init phase {init_phase} --dry-run --config {temp_config}')

            defer.sudo(tmp_dirs_cmd, callback=new_tmp_dirs)

    control_planes.flush()

    stored_manifest = CollectorCallback(cluster)
    generated_manifest = CollectorCallback(cluster)
    for defer in control_planes.get_ordered_members_list():
        old_tmp_dirs_results = old_tmp_dirs.results[defer.get_host()]
        new_tmp_dirs_results = new_tmp_dirs.results[defer.get_host()]
        for i, component in enumerate(components):
            tmp_dir = next(iter(
                set(new_tmp_dirs_results[i].stdout.split())
                - set(old_tmp_dirs_results[i].stdout.split())
            ))
            defer.sudo(f'cat /etc/kubernetes/manifests/{component}.yaml', callback=stored_manifest)
            defer.sudo(f'cat /etc/kubernetes/tmp/{tmp_dir}/{component}.yaml', callback=generated_manifest)

    control_planes.flush()

    result: Dict[str, Dict[str, Optional[str]]] = {}
    for host in control_planes.get_hosts():
        stored_manifest_results = stored_manifest.results[host]
        generated_manifest_results = generated_manifest.results[host]
        for i, component in enumerate(components):
            tofile = (f"{component}.yaml generated from 'services.kubeadm' section"
                      if with_inventory
                      else f"{component}.yaml generated from kubeadm-config ConfigMap")
            stored = stored_manifest_results[i].stdout
            generated = generated_manifest_results[i].stdout
            if component == 'etcd':
                stored = _filter_etcd_initial_cluster_args(stored)
                generated = _filter_etcd_initial_cluster_args(generated)

            diff = utils.get_yaml_diff(stored, generated,
                                       fromfile=f'/etc/kubernetes/manifests/{component}.yaml',
                                       tofile=tofile)

            result.setdefault(host, {})[component] = diff

    return result


def compare_kubelet_config(cluster: KubernetesCluster, *, with_inventory: bool) \
        -> Dict[str, Optional[str]]:
    """
    Generate /var/lib/kubelet/config.yaml in dry-run mode on all nodes,
    and compare with already present configurations.

    :param cluster: KubernetesCluster instance
    :param with_inventory: flag if patches should be taken from the inventory
    :return: mapping host -> diff string
    """
    nodes = cluster.make_group_from_roles(['control-plane', 'worker']).new_defer()
    patches_dir = '/etc/kubernetes/patches'
    if with_inventory:
        patches_dir = utils.get_remote_tmp_path()

    tmp_dirs_cmd = "sh -c 'sudo ls /etc/kubernetes/tmp/ | grep dryrun 2>/dev/null || true'"
    old_tmp_dirs = CollectorCallback(cluster)
    new_tmp_dirs = CollectorCallback(cluster)
    for defer in nodes.get_ordered_members_list():
        if with_inventory:
            defer.sudo(f'mkdir -p {patches_dir}')
            _create_kubeadm_patches_for_component_on_node(
                cluster, defer, 'kubelet', patches_dir=patches_dir, reset=False)

        defer.sudo(tmp_dirs_cmd, callback=old_tmp_dirs)
        defer.sudo(f'kubeadm upgrade node phase kubelet-config --dry-run --patches={patches_dir}')
        defer.sudo(tmp_dirs_cmd, callback=new_tmp_dirs)

    nodes.flush()

    stored_config = CollectorCallback(cluster)
    generated_config = CollectorCallback(cluster)
    for defer in nodes.get_ordered_members_list():
        old_tmp_dirs_results = old_tmp_dirs.results[defer.get_host()]
        new_tmp_dirs_results = new_tmp_dirs.results[defer.get_host()]
        tmp_dir = next(iter(
            set(new_tmp_dirs_results[0].stdout.split())
            - set(old_tmp_dirs_results[0].stdout.split())
        ))
        defer.sudo(f'cat /var/lib/kubelet/config.yaml', callback=stored_config)
        defer.sudo(f'cat /etc/kubernetes/tmp/{tmp_dir}/config.yaml', callback=generated_config)

    nodes.flush()

    result = {}
    for host in nodes.get_hosts():
        tofile = (f"config.yaml with patches from inventory"
                  if with_inventory
                  else f"config.yaml generated from kubelet-config ConfigMap")
        stored = stored_config.results[host][0].stdout
        generated = generated_config.results[host][0].stdout

        diff = utils.get_yaml_diff(stored, generated,
                                   fromfile='/var/lib/kubelet/config.yaml',
                                   tofile=tofile)

        result[host] = diff

    return result


def compare_configmap(cluster: KubernetesCluster, configmap: str) -> Optional[str]:
    control_plane = cluster.nodes['control-plane'].get_first_member()
    kubeadm_config = KubeadmConfig(cluster)

    if configmap == 'kubelet-config':
        # Do not check kubelet-config ConfigMap, because some properties may be deleted from KubeletConfiguration
        # if set to default, for example readOnlyPort: 0, protectKernelDefaults: false
        # Otherwise, the check would require to take into account all such default properties.
        if not kubeadm_extended_dryrun(cluster):
            return None

        # Use upload-config kubelet --dry-run to catch all inserted/updated/deleted properties.

        temp_config = utils.get_remote_tmp_path()
        patches_dir = utils.get_remote_tmp_path()

        defer = control_plane.new_defer()
        collector = CollectorCallback(cluster)

        _upload_config(cluster, defer, kubeadm_config, temp_config, patches_dir=patches_dir)
        defer.sudo(f'mkdir -p {patches_dir}')
        _create_kubeadm_patches_for_component_on_node(
            cluster, defer, 'kubelet', patches_dir=patches_dir, reset=False)

        init_phase = CONFIGMAPS_CONSTANTS[configmap]['init_phase']
        defer.sudo(f'kubeadm init phase {init_phase} --dry-run --config {temp_config}',
                   callback=collector)

        defer.flush()
        output = collector.result[control_plane.get_host()].stdout

        split_logs = re.compile(r'^\[.*].*$\n', flags=re.M)
        cfg = next(filter(lambda ln: 'kind: KubeletConfiguration' in ln, split_logs.split(output)))
        cfg = dedent(cfg)

        key = CONFIGMAPS_CONSTANTS[configmap]['key']
        generated_config = yaml.safe_load(cfg)['data'][key]
        if 'resolvConf' not in kubeadm_config.maps[configmap]:
            generated_config = _filter_kubelet_configmap_resolv_conf(generated_config)

        kubeadm_config.load(configmap, control_plane)
        # Use loaded_maps that preserve original formatting
        stored_config = kubeadm_config.loaded_maps[configmap].obj["data"][key]

        if yaml.safe_load(generated_config) == yaml.safe_load(stored_config):
            return None

        return utils.get_unified_diff(stored_config, generated_config,
                                      fromfile=f'{configmap} ConfigMap',
                                      tofile="generated from 'services.kubeadm_kubelet' section")

    else:
        # Merge with inventory and check.
        # This way it is possible to check only new or changed properties in the inventory
        # that are still not reflected in the remote ConfigMap.

        stored_config = kubeadm_config.load(configmap, control_plane)

        generated_config = kubeadm_config.merge_with_inventory(configmap)\
                (utils.deepcopy_yaml(stored_config))

        if generated_config == stored_config:
            return None

        section = CONFIGMAPS_CONSTANTS[configmap]['section']
        return utils.get_unified_diff(yaml.dump(stored_config), yaml.dump(generated_config),
                                      fromfile=f'{configmap} ConfigMap',
                                      tofile=f"{configmap} ConfigMap merged 'services.{section}' section")


def _detect_changes(logger: log.EnhancedLogger, old: str, new: str, fromfile: str, tofile: str) -> bool:
    diff = utils.get_yaml_diff(old, new, fromfile, tofile)
    if diff is not None:
        logger.debug(f"Detected changes in {tofile}")
        logger.verbose(diff)
        return True

    return False


def _filter_etcd_initial_cluster_args(content: str) -> str:
    return '\n'.join(ln for ln in content.splitlines() if '--initial-cluster' not in ln)


def _filter_kubelet_configmap_resolv_conf(content: str) -> str:
    return '\n'.join(ln for ln in content.splitlines() if 'resolvConf:' not in ln)


def _restart_containers(cluster: KubernetesCluster, node: NodeGroup, components: Sequence[str]) -> None:
    if not components:
        return

    logger = cluster.log
    node_name = node.get_node_name()
    logger.debug(f"Restarting containers for components {list(components)} on node: {node_name}")

    commands = []

    # Take into account probably missed container because kubelet may be restarting them at this moment.
    # Though still ensure the command to delete the container successfully if it is present.
    restart_container = ("(set -o pipefail && sudo crictl ps --name {component} -q "
                         "| xargs -I CONTAINER sudo crictl rm -f CONTAINER) > /dev/null")

    for component in components:
        commands.append(restart_container.format(component=component))

    get_container_from_cri = "sudo crictl ps --name {component} -q"
    get_container_from_pod = (
        "sudo kubectl get pods -n kube-system {component}-{node} "
        "-o 'jsonpath={{.status.containerStatuses[0].containerID}}{{\"\\n\"}}' "
        "| sed 's|.\\+://\\(.\\+\\)|\\1|'")

    # Wait for kubelet to refresh container status in pods.
    # It is expected that Ready status will be refreshed at the same time,
    # so we can safely wait_for_pods().
    test_refreshed_container = (
        f"("
        f"CONTAINER=$({get_container_from_cri}); "
        f"if [ -z \"$CONTAINER\" ]; then "
        f"  echo \"container '{{component}}' is not created yet\" >&2 ; exit 1; "
        f"fi "
        f"&& "
        f"if [ \"$CONTAINER\" != \"$({get_container_from_pod})\" ]; "
        f"  then echo \"Pod '{{component}}-{{node}}' is not refreshed yet\" >&2; exit 1; "
        f"fi "
        f")")

    for component in components:
        commands.append(test_refreshed_container.format(component=component, node=node_name))

    expect_config = cluster.inventory['globals']['expect']['pods']['kubernetes']
    node.wait_commands_successful(commands,
                                  timeout=expect_config['timeout'],
                                  retries=expect_config['retries'],
                                  sudo=False, pty=True)


def _delete_pods(cluster: KubernetesCluster, node: AbstractGroup[RunResult],
                 control_plane: NodeGroup, components: Sequence[str]) -> None:
    if not components:
        return

    node_name = node.get_node_name()
    cluster.log.debug(f"Deleting pods for components {list(components)} on node: {node_name}")

    restart_pod = (
        "kubectl delete pod -n kube-system $("
        "    sudo kubectl get pods -n kube-system -o=wide "
        "    | grep '{pod}' | grep '{node}' | awk '{{ print $1 }}'"
        ")"
    )

    defer = control_plane.new_defer()
    for component in components:
        defer.sudo(restart_pod.format(pod=component, node=node_name))

    defer.flush()


# function to get dictionary of flags to be patched for a given control plane item and a given node
def get_patched_flags_for_section(cluster: KubernetesCluster,
                                  patch_section: str, group: AbstractGroup[RunResult]) -> Dict[str, Union[str, bool, int]]:
    node = group.get_config()
    flags = {}

    for n in cluster.inventory['services']['kubeadm_patches'][patch_section]:
        group = cluster.create_group_from_groups_nodes_names(
            n.get('groups', []), n.get('nodes', []))

        if group.has_node(node['name']):
            for arg, value in n['patch'].items():
                flags[arg] = value

    # we always set binding-address to the node's internal address for apiServer
    if patch_section == 'apiServer' and 'control-plane' in node['roles']:
        flags['bind-address'] = node['internal_address']

    return flags


def _create_kubeadm_patches_for_component_on_node(cluster: KubernetesCluster, node: DeferredGroup, component: str,
                                                  backup_dir: Optional[str] = None,
                                                  *, patches_dir: str = '/etc/kubernetes/patches', reset: bool = True) -> None:
    patch_constants = COMPONENTS_CONSTANTS[component]['patch']
    component_patches = f"find {patches_dir} -name {patch_constants['target_template']}"

    if backup_dir is not None:
        node.sudo(f'{component_patches} -exec cp {{}} {backup_dir}/patches \\;')

    if reset:
        node.sudo(f'{component_patches} -delete')

    # read patch content from inventory and upload patch files to a node
    patched_flags = get_patched_flags_for_section(cluster, patch_constants['section'], node)
    if patched_flags:
        if component == 'kubelet':
            template_filename = 'templates/patches/kubelet.yaml.j2'
        else:
            template_filename = 'templates/patches/control-plane-pod.json.j2'

        control_plane_patch = Template(utils.read_internal(template_filename)).render(flags=patched_flags)
        patch_file = patches_dir + '/' + patch_constants['file']
        node.put(io.StringIO(control_plane_patch + "\n"), patch_file, sudo=True)
        node.sudo(f'chmod 644 {patch_file}')



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

from kubemarine import kubernetes
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup


def k8s_certs_overview(control_planes: NodeGroup) -> None:
    for control_plane in control_planes.get_ordered_members_list():
        control_planes.cluster.log.debug(f"Checking certs expiration for control_plane {control_plane.get_node_name()}")
        control_plane.sudo("kubeadm certs check-expiration", hide=False, pty=True)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['cert_renew'])
def renew_verify(cluster: KubernetesCluster) -> None:
    if "kubernetes" not in cluster.procedure_inventory:
        return

    cert_list = cluster.procedure_inventory["kubernetes"].get("cert-list")
    verify_all_is_absent_or_single(cert_list)


def renew_apply(control_planes: NodeGroup) -> None:
    log = control_planes.cluster.log

    procedure = control_planes.cluster.procedure_inventory["kubernetes"]
    cert_list = procedure["cert-list"]

    for cert in cert_list:
        control_planes.sudo(f"kubeadm certs renew {cert}", pty=True)

    if "all" in cert_list or "admin.conf" in cert_list:
        # need to update cluster-admin config
        kubernetes.copy_admin_config(log, control_planes)

    # for some reason simple pod delete do not work for certs update - we need to delete containers themselves
    control_planes.call(kubernetes.components.restart_components,
                        components=kubernetes.components.CONTROL_PLANE_COMPONENTS)


def verify_all_is_absent_or_single(cert_list: List[str]) -> None:
    if "all" in cert_list and len(cert_list) > 1:
        raise Exception(f"Found 'all' in certs list, but it is not single: {cert_list}")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import sys
import traceback
from abc import ABC, abstractmethod

from textwrap import dedent
from typing import Type, List

from kubemarine.core import log as klog


class GroupException(Exception):
    pass


def get_kme_dictionary() -> dict:
    return {
        "KME0000": {
            "name": "Test exception"
        },
        "KME0001": {
            "instance": KeyboardInterrupt,
            "name": "Interrupted{reason}"
        },
        "KME0002": {
            "instance": GroupException,
            "name": "Remote group exception\n{reason}"
        },
        "KME0004": {
            "name": "There are no control planes defined in the cluster scheme"
        },
        "KME0005": {
            "name": "{hostnames} are not sudoers"
        },
        "KME0006": {
            "offline": dedent(
                """\
                Nodes {offline} are not reachable.
                Check that the nodes addresses are entered correctly in the inventory.
                The nodes may also be turned off, SSH port is probably not opened or incorrect,
                or something is incorrect with the SSH daemon.
                """.rstrip()
            ),
            "inaccessible": dedent(
                """\
                Nodes {inaccessible} are not accessible through SSH.
                Check that the SSH credentials (keyfile, username, password) are entered correctly in the inventory.
                """.rstrip()
            )
        },
        "KME0008": {
            "name": "Specified Kubernetes version '{version}' - cannot be used! "
                    "Allowed versions are: {allowed_versions}."
        },
        "KME0009": {
            "name": "Key {key!r} is redefined for {plugin_name!r} in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required plugin configuration explicitly in procedure inventory."
        },
        "KME0010": {
            "name": "Associations are redefined for {package!r} in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required associations explicitly in procedure inventory."
        },
        "KME0011": {
            "name": "Key {key!r} is redefined for third-party {thirdparty!r} in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required third-party configuration explicitly in procedure inventory."
        },
        "KME0012": {
            "name": "Procedure {procedure!r} is possible only for cluster "
                    "with all nodes having the same and supported OS family."
        },
        'KME0013': {
            "name": "Key 'plugins.\"io.containerd.grpc.v1.cri\".sandbox_image' is redefined for 'containerdConfig' "
                    "in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required 'sandbox_image' explicitly in procedure inventory."
        }
    }


class BaseKME(RuntimeError, ABC):
    def __init__(self, code: str):
        self.code = code
        if self.code not in get_kme_dictionary():
            raise ValueError('An error was raised with an unknown error code')
        self.kme: dict = get_kme_dictionary()[self.code]
        self.message = self._format()
        super().__init__(self.message)

    @abstractmethod
    def _format(self) -> str: ...

    def __str__(self) -> str:
        return self.code + ": " + self.message


class KME(BaseKME):
    def __init__(self, code: str, **kwargs: object):
        self.kwargs = kwargs
        super().__init__(code)

    def _format(self) -> str:
        if 'name' not in self.kme:
            raise ValueError('An error was raised with an unsupported error code')

        name: str = self.kme['name']
        return name.format_map(self.kwargs)


class KME0006(BaseKME):
    def __init__(self, offline: List[str], inaccessible: List[str]):
        self.offline = offline
        self.inaccessible = inaccessible
        self.summary = ""
        self.details = ""
        super().__init__("KME0006")

    def _format(self) -> str:
        self.summary = f"Failed to connect to {len(self.offline + self.inaccessible)} nodes."
        msgs = []
        if self.offline:
            msgs.append(self.kme['offline'].format(offline=self.offline))
        if self.inaccessible:
            msgs.append(self.kme['inaccessible'].format(inaccessible=self.inaccessible))
        self.details = '\n'.join(msgs)
        return self.summary + '\n' + self.details


class FailException(Exception):
    def __init__(self, message: str = '', reason: BaseException = None, hint: str = '') -> None:
        super().__init__(message)
        self.message = message
        self.reason = reason
        self.hint = hint


def wrap_kme_exception(reason: BaseException) -> BaseException:
    for dictionary_code, dictionary_kme in get_kme_dictionary().items():
        exception_class: Type[BaseException] = dictionary_kme.get('instance')
        if exception_class is not None and isinstance(reason, exception_class):
            return KME(dictionary_code, reason=reason)

    return reason


def pretty_print_error(message: str = '', reason: BaseException = None, log: klog.EnhancedLogger = None) -> None:
    """
    Parses the passed error and nicely displays its name and structure depending on what was passed.
    The method outputs to stdout by default, but will use the logger if one is specified.

    :param message: an optional message describing context of the error
    :param reason: an exception that caused the failure.
    :param log: logger object, if you need to write a log there
    :return: None
    """
    def error_logger(msg: object) -> None:
        if log:
            log.critical(msg)
        else:
            sys.stderr.write(str(msg) + "\n")

    error_logger('FAILURE!')
    if message != '':
        error_logger(message)

    if reason is None:
        return

    reason = wrap_kme_exception(reason)

    if isinstance(reason, BaseKME):
        error_logger(reason)
        return

    if log:
        log.critical('KME0001: Unexpected exception', exc_info=reason)
    else:
        sys.stderr.write("KME0001: Unexpected exception\n")
        exc_info = (type(reason), reason, reason.__traceback__)
        sio = io.StringIO()
        traceback.print_exception(*exc_info, limit=None, file=sio)
        sys.stderr.write(sio.getvalue())



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import List, Dict, Tuple, Optional, Union, Mapping, Set, Protocol
from io import StringIO

from kubemarine import yum, apt, jinja
from kubemarine.core import errors, static, utils
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.executor import RunnersResult, Token, Callback
from kubemarine.core.group import (
    NodeGroup, DeferredGroup, AbstractGroup, RunResult, GROUP_RUN_TYPE,
    CollectorCallback, RunnersGroupResult
)
from kubemarine.core.yaml_merger import default_merger


ERROR_GLOBAL_ASSOCIATIONS_REDEFINED_MULTIPLE_OS = \
    "It is not supported to customize services.packages.associations section " \
    "if nodes have different OS families. " \
    "Please move the section to corresponding services.packages.associations.<os_family> section."

ERROR_MULTIPLE_PACKAGE_VERSIONS_DETECTED = \
    "Multiple package versions detected %s for package '%s'. " \
    "Align them to the single version manually or using corresponding task of install procedure. " \
    "Alternatively, specify cache_versions=false for corresponding association."

ERROR_SEMANAGE_NOT_MANAGED_DEBIAN = "semanage is not managed for debian OS family by KubeMarine"


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    enrich_inventory_associations(cluster)
    enrich_inventory_packages(cluster)


def enrich_inventory_associations(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    kubernetes_version = inventory['services']['kubeadm']['kubernetesVersion']
    associations: dict = inventory['services']['packages']['associations']
    inventory['services']['packages']['associations'] = enriched_associations = {}

    # Move associations for OS families and merge with globals
    for association_name in get_associations_os_family_keys():
        redefined_associations = associations.pop(association_name)
        if cluster.nodes['all'].get_subgroup_with_os(association_name).is_empty():
            continue

        os_associations = utils.deepcopy_yaml(static.GLOBALS['packages']['common_associations'])
        if association_name == 'debian':
            del os_associations['semanage']
        for association_params in os_associations.values():
            del association_params['groups']

        for package in static.GLOBALS['packages'][association_name]:
            os_associations[package]['package_name']\
                = get_default_package_names(association_name, package, kubernetes_version)

        default_merger.merge(os_associations, redefined_associations)
        enriched_associations[association_name] = os_associations

    # Check remained associations section if they are customized at global level.
    if associations:
        os_family = cluster.get_os_family()
        if os_family == 'multiple':
            raise Exception(ERROR_GLOBAL_ASSOCIATIONS_REDEFINED_MULTIPLE_OS)
        elif os_family not in ('unknown', 'unsupported', '<undefined>'):
            # move remained associations properties to the specific OS family section and merge with priority
            default_merger.merge(enriched_associations[os_family], associations)

    if 'semanage' in enriched_associations.get('debian', {}):
        raise Exception(ERROR_SEMANAGE_NOT_MANAGED_DEBIAN)


def enrich_inventory_packages(cluster: KubernetesCluster) -> None:
    for _type in ['install', 'upgrade', 'remove']:
        packages_list = cluster.inventory['services']['packages'].get(_type)
        if isinstance(packages_list, list):
            cluster.inventory['services']['packages'][_type] = {
                'include': packages_list
            }

    for _type in ['upgrade', 'remove']:
        packages: dict = cluster.inventory['services']['packages'].get(_type)
        if packages is not None:
            packages.setdefault('include', ['*'])


def _get_associations_procedure_plan(cluster: KubernetesCluster) -> List[Tuple[str, dict]]:
    context = cluster.context
    procedure = context["initial_procedure"]
    procedure_inventory = cluster.procedure_inventory
    upgrade_plan = []
    if procedure == "upgrade":
        kubernetes_version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
        for i, v in enumerate(procedure_inventory['upgrade_plan'][context['upgrade_step']:]):
            # Take the target (probably) compiled version and the remained not yet compiled
            version = kubernetes_version if i == 0 else v
            procedure_associations = procedure_inventory.get(v, {}).get("packages", {}).get("associations", {})
            procedure_associations = utils.subdict_yaml(procedure_associations,
                                                        _get_system_packages_support_upgrade(cluster))

            upgrade_plan.append((version, procedure_associations))

    elif procedure == "migrate_kubemarine" and "upgrading_package" in context:
        procedure_associations = procedure_inventory.get('upgrade', {}).get("packages", {}).get("associations", {})
        procedure_associations = utils.subdict_yaml(procedure_associations,
                                                    _get_system_packages_support_upgrade(cluster))
        upgrade_plan = [("", procedure_associations)]

    return upgrade_plan


def _get_redefined_package_name(cluster: KubernetesCluster, associations: dict, package: str) \
        -> Optional[List[str]]:
    # Global section has priority
    package_name: Union[str, List[str], None] = associations.get(package, {}).get('package_name')
    if not package_name:
        package_name = associations.get(cluster.get_os_family(), {}).get(package, {}).get('package_name')

    return [package_name] if isinstance(package_name, str) else package_name


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade', 'migrate_kubemarine'])
def enrich_procedure_inventory(cluster: KubernetesCluster) -> None:
    procedure_plan = _get_associations_procedure_plan(cluster)
    procedure_associations = {} if not procedure_plan else procedure_plan[0][1]

    if procedure_associations:
        # Merge global associations section because it has priority during enrichment.
        cluster_associations = cluster.inventory.setdefault("services", {}).setdefault("packages", {}) \
            .setdefault("associations", {})
        default_merger.merge(cluster_associations, utils.deepcopy_yaml(procedure_associations))

    if cluster.context['initial_procedure'] == 'upgrade':
        upgrade_inventory_packages(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade', 'migrate_kubemarine',])
def verify_procedure_inventory(cluster: KubernetesCluster) -> None:
    context = cluster.context
    procedure = context["initial_procedure"]

    os_family = cluster.get_os_family()
    if os_family not in get_associations_os_family_keys():
        raise errors.KME("KME0012", procedure=procedure)

    upgrade_plan = _get_associations_procedure_plan(cluster)
    if not upgrade_plan:
        return

    if procedure == "upgrade":
        previous_version = cluster.previous_inventory['services']['kubeadm']['kubernetesVersion']
    else:  # migrate_kubemarine procedure
        previous_version = ""

    packages_verify = _get_system_packages_support_upgrade(cluster)
    _verify_upgrade_plan(cluster, previous_version, packages_verify, upgrade_plan)


def upgrade_inventory_packages(cluster: KubernetesCluster) -> None:
    procedure_inventory = cluster.procedure_inventory
    upgrade_version = cluster.procedure_inventory['upgrade_plan'][cluster.context['upgrade_step']]
    for _type in ['install', 'upgrade', 'remove']:
        packages_section = procedure_inventory.get(upgrade_version, {}).get("packages", {})
        upgrade_packages = packages_section.get(_type)
        if upgrade_packages is None:
            continue
        if isinstance(upgrade_packages, list):
            upgrade_packages = {'include': upgrade_packages}

        packages_section = cluster.inventory.setdefault("services", {}).setdefault("packages", {})
        inventory_packages = packages_section.setdefault(_type, {})
        if isinstance(inventory_packages, list):
            packages_section[_type] = inventory_packages = {
                'include': inventory_packages
            }

        default_merger.merge(inventory_packages, upgrade_packages)


def _verify_upgrade_plan(cluster: KubernetesCluster, previous_version: str,
                         packages_verify: List[str], upgrade_plan: List[Tuple[str, dict]]) -> None:
    raw_associations = utils.deepcopy_yaml(cluster.previous_raw_inventory.get('services', {}).get('packages', {})
                                           .get('associations', {}))

    # validate all packages sections in procedure inventory
    for version, upgrade_associations in upgrade_plan:
        if version != '' and jinja.is_template(version):
            break

        for package in packages_verify:
            upgrade_package_name = upgrade_associations.get(package, {}).get('package_name')
            if _get_redefined_package_name(cluster, raw_associations, package) and not upgrade_package_name:
                raise errors.KME("KME0010", package=package,
                                 previous_version_spec=f' for version {previous_version}' if previous_version else "",
                                 next_version_spec=f' for version {version}' if version else "")

        default_merger.merge(raw_associations, utils.deepcopy_yaml(upgrade_associations))
        previous_version = version


def get_default_package_names(os_family: str, package: str, kubernetes_version: str) -> List[str]:
    version_key = get_compatibility_version_key(os_family)
    compatibility = static.GLOBALS['compatibility_map']['software']
    packages_names: List[Dict[str, str]] = static.GLOBALS['packages'][os_family][package]['package_name']

    package_versions = []
    for kv in packages_names:
        package_name, software_name = next((k, v) for k, v in kv.items())
        if software_name in ('haproxy', 'keepalived'):
            version = compatibility[software_name][version_key]
        else:
            version = compatibility[software_name][kubernetes_version][version_key]

        package_versions.append(f"{package_name}{get_package_version_separator(os_family)}{version}")

    return package_versions


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade', 'migrate_kubemarine'])
def calculate_upgrade_required(cluster: KubernetesCluster) -> None:
    os_family = cluster.get_os_family()

    context = cluster.context
    system_packages = _get_system_packages_support_upgrade(cluster)
    upgrade_required = context.setdefault("upgrade", {}).setdefault('required', {}).setdefault('packages', [])
    for package in system_packages:
        if (cluster.previous_inventory["services"]["packages"]["associations"][os_family][package]['package_name']
                != cluster.inventory["services"]["packages"]["associations"][os_family][package]['package_name']):
            upgrade_required.append(package)
            continue

        raw_associations = (cluster.previous_raw_inventory.get('services', {}).get('packages', {})
                            .get('associations', {}))
        raw_package_name = _get_redefined_package_name(cluster, raw_associations, package)
        if (context['initial_procedure'] == 'migrate_kubemarine'
                and (not raw_package_name or any(jinja.is_template(pkg) for pkg in raw_package_name))):
            # If package name is redefined with template,
            # upgrade may be required as we have lost previous compilation result.
            upgrade_required.append(package)


def _get_system_packages_support_upgrade(cluster: KubernetesCluster) -> List[str]:
    context = cluster.context
    procedure = context["initial_procedure"]
    if procedure == 'upgrade':
        return ['containerd']
    elif procedure == "migrate_kubemarine" and "upgrading_package" in context:
        return [context['upgrading_package']]

    return []


def cache_package_versions(cluster: KubernetesCluster, inventory: dict, by_initial_nodes: bool = False) -> dict:
    if cluster.get_os_family() == '<undefined>':
        # All nodes are inaccessible. This is possible only in check_iaas.
        return inventory

    os_ids = cluster.get_os_identifiers()
    different_os = list(set(os_ids.values()))
    if len(different_os) > 1:
        cluster.log.debug(f"Final nodes have different OS families or versions, packages will not be cached. "
                          f"List of (OS family, version): {different_os}")
        return inventory

    os_family = different_os[0][0]
    if os_family in ('unknown', 'unsupported'):
        # For add_node/install procedures we check that OS is supported in prepare.check.system task.
        # For check_iaas procedure it is allowed to have unsupported OS, so skip caching.
        cluster.log.debug("Skip caching of packages for unsupported OS.")
        return inventory

    group = (cluster.previous_nodes if by_initial_nodes else cluster.nodes)['all'] \
        .exclude_group(cluster.nodes['all'].get_online_nodes(False))

    if group.nodes_amount() != group.get_sudo_nodes().nodes_amount():
        # For add_node/install procedures we check that all nodes are sudoers in prepare.check.sudoer task.
        # For check_iaas procedure the nodes might still be not sudoers.
        # Skip caching if any not-sudoer node found.
        cluster.log.debug(f"Some nodes are not sudoers, packages will not be cached.")
        return inventory

    hosts_to_packages = get_all_managed_packages_for_group(group, inventory, by_initial_nodes)
    detected_packages = detect_installed_packages_version_hosts(cluster, hosts_to_packages)

    _cache_package_associations(group, inventory, detected_packages, by_initial_nodes)
    _cache_custom_packages(cluster, inventory, detected_packages, by_initial_nodes)

    cluster.log.debug('Package versions detection finished')
    return inventory


def get_all_managed_packages_for_group(group: NodeGroup, inventory: dict, ensured_association_only: bool = False) \
        -> Dict[str, List[str]]:
    """
    Returns hosts with list of all managed packages for them.
    For associations, only subset of hosts is considered on which the associations are managed by KubeMarine.

    :param group: Group of nodes to get the manager packages for.
    :param inventory: Inventory of the cluster. May be different from the inventory of the cluster instance,
                      if used during finalization.
    :param ensured_association_only: Specify whether to take 'cache_versions' property into account for associations.
                                     Additionally, if true, will skip custom packages.
    :return: List of packages for each relevant host.
    """
    packages_section = inventory['services']['packages']
    hosts_to_packages: Dict[str, List[str]] = {}
    for node in group.get_ordered_members_list():
        os_family = node.get_nodes_os()
        node_associations = packages_section['associations'].get(os_family, {})
        for association_name in node_associations.keys():
            host_packages = get_association_hosts_to_packages(
                node, inventory, association_name, ensured_association_only)

            packages: List[str] = next(iter(host_packages.values()), [])
            hosts_to_packages.setdefault(node.get_host(), []).extend(packages)

    custom_install_packages = inventory['services']['packages'].get('install', {}).get('include', [])
    if not ensured_association_only and custom_install_packages:
        for host in group.get_hosts():
            hosts_to_packages.setdefault(host, []).extend(custom_install_packages)

    return hosts_to_packages


def get_association_hosts_to_packages(group: AbstractGroup[RunResult], inventory: dict, association_name: str,
                                      ensured_association_only: bool = False) \
        -> Dict[str, List[str]]:
    """
    Returns hosts with associated packages list for the specified association name.
    Only subset of hosts is returned on which the association is managed by KubeMarine.

    :param group: Group of nodes to check the applicability of the association.
    :param inventory: Inventory of the cluster. May be different from the inventory of the cluster instance,
                      if used during finalization.
    :param association_name: target association name
    :param ensured_association_only: Specify whether to take 'cache_versions' property into account.
    :return: List of packages for each relevant host.
    """
    cluster: KubernetesCluster = group.cluster

    packages_section = inventory['services']['packages']
    if not packages_section['mandatory'].get(association_name, True):
        return {}

    hosts_to_packages = {}

    if association_name == 'unzip':
        from kubemarine import thirdparties
        relevant_group: AbstractGroup[RunResult] = thirdparties.get_group_require_unzip(cluster, inventory)
    else:
        groups = cluster.globals['packages']['common_associations'].get(association_name, {}).get('groups', [])
        relevant_group = cluster.make_group_from_roles(groups)

    relevant_group = relevant_group.intersection_group(group)

    for node in relevant_group.get_ordered_members_list():
        os_family = node.get_nodes_os()
        packages = get_association_packages(cluster, os_family, association_name)

        if ensured_association_only and not cache_versions_enabled(cluster, os_family, association_name):
            packages = []

        if packages:
            hosts_to_packages[node.get_host()] = packages

    return hosts_to_packages


def get_association_packages(cluster: KubernetesCluster, os_family: str, association_name: str) -> List[str]:
    packages_section = cluster.inventory['services']['packages']
    package_associations = packages_section['associations'].get(os_family, {}).get(association_name, {})
    packages: Union[str, List[str]] = package_associations.get('package_name', [])

    if isinstance(packages, str):
        packages = [packages]

    return packages


def cache_versions_enabled(cluster: KubernetesCluster, os_family: str, association_name: str) -> bool:
    packages_section = cluster.inventory['services']['packages']
    global_cache_versions = packages_section['cache_versions']
    specific_cache_versions: bool = packages_section['associations'] \
        .get(os_family, {}).get(association_name, {}).get('cache_versions', True)

    return global_cache_versions and specific_cache_versions


def _cache_package_associations(group: NodeGroup, inventory: dict,
                                detected_packages: Dict[str, Dict[str, List]], ensured_association_only: bool) -> None:
    cluster: KubernetesCluster = group.cluster
    associations = inventory['services']['packages']['associations'][cluster.get_os_family()]
    for association_name, associated_params in associations.items():
        hosts_to_packages = get_association_hosts_to_packages(
            group, inventory, association_name, ensured_association_only)
        if not hosts_to_packages:
            continue

        # Since all nodes have the same OS family in this case,
        # the packages list is the same for all relevant hosts, so take any available.
        packages_list = next(iter(hosts_to_packages.values()))

        final_packages_list = []
        for package in packages_list:
            final_package = _detect_final_package(cluster, detected_packages, package, ensured_association_only)
            final_packages_list.append(final_package)

        # if non-multiple value, then convert to simple string
        # packages can contain multiple package values
        if len(final_packages_list) == 1:
            associated_params['package_name'] = final_packages_list[0]
        else:
            associated_params['package_name'] = final_packages_list


def _cache_custom_packages(cluster: KubernetesCluster, inventory: dict,
                           detected_packages: Dict[str, Dict[str, List]], ensured_association_only: bool) -> None:
    if ensured_association_only:
        return
    # packages from direct installation section
    custom_install_packages = inventory['services']['packages'].get('install', {})
    if custom_install_packages.get('include', []):
        final_packages_list = []
        for package in custom_install_packages['include']:
            final_package = _detect_final_package(cluster, detected_packages, package, False)
            final_packages_list.append(final_package)
        custom_install_packages['include'] = final_packages_list


def _detect_final_package(cluster: KubernetesCluster, detected_packages: Dict[str, Dict[str, List]],
                          package: str, ensured_association_only: bool) -> str:
    # add package version to list only if it was found as installed
    detected_package_versions = list(filter(lambda version: "not installed" not in version,
                                            detected_packages[package].keys()))

    # if there no versions detected, then return default package from inventory
    if not detected_package_versions:
        return package
    elif len(detected_package_versions) > 1:
        if ensured_association_only:
            raise Exception(ERROR_MULTIPLE_PACKAGE_VERSIONS_DETECTED % (str(detected_packages[package]), package))
        else:
            cluster.log.warning(
                f"Multiple package versions detected {detected_packages[package]} for package '{package}'. "
                f"Use default package '{package}' from inventory.")
            # return default package from inventory if multiple versions detected
            return package
    else:
        return detected_package_versions[0]


def disable_unattended_upgrade(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    if group.get_nodes_os() != 'debian':
        cluster.log.debug("Skipped - unattended upgrades are supported only on Ubuntu/Debian os family")
        return

    packages_per_node = get_all_managed_packages_for_group(group=group, inventory=cluster.inventory,
                                                           ensured_association_only=True)

    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            packages = [get_package_name(node.get_nodes_os(), package) for package in packages_per_node[node.get_host()]]
            unattended_upgrade_config = 'Unattended-Upgrade::Package-Blacklist { %s };\n' % " ".join(
                ['"%s";' % package for package in packages])
            node.put(StringIO(unattended_upgrade_config), '/etc/apt/apt.conf.d/51unattended-upgrades-kubemarine',
                     sudo=True, backup=True)


def get_associations_os_family_keys() -> Set[str]:
    return {'debian', 'rhel', 'rhel8', 'rhel9'}


def get_compatibility_version_key(os_family: str) -> str:
    """
    Get os-specific version key to be used in software compatibility map.
    :param os_family: one of supported OS families
    :return: String to use as version key.
    """
    if os_family in get_associations_os_family_keys():
        return f"version_{os_family}"
    else:
        raise ValueError(f"Unsupported {os_family!r} OS family")


class PackageManager(Protocol):
    def ls_repofiles(self, group: NodeGroup) -> RunnersGroupResult: ...

    def backup_repo(self, group: NodeGroup) -> Optional[RunnersGroupResult]: ...

    def add_repo(self, group: NodeGroup, repo_data: Union[List[str], Dict[str, dict], str]) -> RunnersGroupResult: ...

    def get_repo_file_name(self) -> str: ...

    def create_repo_file(self, group: AbstractGroup[RunResult],
                         repo_data: Union[List[str], Dict[str, dict], str],
                         repo_file: str) -> None: ...

    def clean(self, group: NodeGroup) -> RunnersGroupResult: ...

    def get_install_cmd(self, include: Union[str, List[str]], exclude: Union[str, List[str]] = None) -> str: ...

    def install(self, group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
                exclude: Union[str, List[str]] = None,
                pty: bool = False, callback: Callback = None) -> GROUP_RUN_TYPE: ...

    def remove(self, group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
               exclude: Union[str, List[str]] = None,
               warn: bool = False, hide: bool = True, pty: bool = False) -> GROUP_RUN_TYPE: ...

    def upgrade(self, group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
                exclude: Union[str, List[str]] = None,
                pty: bool = False) -> GROUP_RUN_TYPE: ...

    def no_changes_found(self, action: str, result: RunnersResult) -> bool: ...

    def search(self, group: DeferredGroup, package: str, callback: Callback = None) -> Token: ...


def get_package_manager(group: AbstractGroup[GROUP_RUN_TYPE]) -> PackageManager:
    os_family = group.get_nodes_os()

    if os_family in ['rhel', 'rhel8', 'rhel9']:
        return yum
    elif os_family == 'debian':
        return apt

    raise Exception('Failed to return package manager for unknown or multiple OS')


def ls_repofiles(group: NodeGroup) -> RunnersGroupResult:
    return get_package_manager(group).ls_repofiles(group)


def backup_repo(group: NodeGroup) -> Optional[RunnersGroupResult]:
    return get_package_manager(group).backup_repo(group)


def add_repo(group: NodeGroup, repo_data: Union[List[str], dict, str]) -> RunnersGroupResult:
    return get_package_manager(group).add_repo(group, repo_data)


def get_repo_filename(group: AbstractGroup[RunResult]) -> str:
    return get_package_manager(group).get_repo_file_name()


def create_repo_file(group: AbstractGroup[RunResult], repo_data: Union[List[str], dict, str], repo_file: str) -> None:
    get_package_manager(group).create_repo_file(group, repo_data, repo_file)


def clean(group: NodeGroup) -> RunnersGroupResult:
    return get_package_manager(group).clean(group)


def install(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None,
            pty: bool = False, callback: Callback = None) -> GROUP_RUN_TYPE:
    return get_package_manager(group).install(group, include, exclude,
                                              pty=pty, callback=callback)


def remove(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None, exclude: Union[str, List[str]] = None,
           warn: bool = False, hide: bool = True, pty: bool = False) -> GROUP_RUN_TYPE:
    return get_package_manager(group).remove(group, include, exclude, warn=warn, hide=hide, pty=pty)


def upgrade(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None,
            pty: bool = False) -> GROUP_RUN_TYPE:
    return get_package_manager(group).upgrade(group, include, exclude, pty=pty)


def no_changes_found(group: NodeGroup, action: str, result: RunnersResult) -> bool:
    pkg_mgr = get_package_manager(group)
    return pkg_mgr.no_changes_found(action, result)


def search_package(group: DeferredGroup, package: str, callback: Callback = None) -> Token:
    return get_package_manager(group).search(group, package, callback)


def get_detect_package_version_cmd(os_family: str, package_name: str) -> str:
    if os_family in ["rhel", "rhel8", "rhel9"]:
        cmd = r"rpm -q %s" % package_name
    else:
        cmd = r"dpkg-query -f '${Package}=${Version}\n' -W %s" % package_name

    return cmd


def _detect_installed_package_version(group: DeferredGroup, package: str, collector: CollectorCallback) -> Token:
    """
    Detect package versions for each host on remote group
    :param group: Group of nodes, where package should be found
    :param package: package name, which version should be detected (eg. 'containerd')
    :return: NodeGroupResults with package version on each host

    Method generates different package query for different OS.

    Note: for Ubuntu/Debian some packages returns multiline results for some queries
    (for example docker-ce* returns docker-ce and docker-ce-cli).
    """

    os_family = group.get_nodes_os()
    package_name = get_package_name(os_family, package)

    cmd = get_detect_package_version_cmd(os_family, package_name)
    return group.sudo(cmd, warn=True, callback=collector)


def _parse_node_detected_package(result: RunnersResult, package: str) -> str:
    node_detected_package = result.stdout.strip() + result.stderr.strip()
    # consider version, which ended with special symbol = or - as not installed
    # (it is possible in some cases to receive "containerd=" version)
    if "not installed" in node_detected_package or "no packages found" in node_detected_package \
            or node_detected_package[-1] == '=' or node_detected_package[-1] == '-':
        node_detected_package = f"not installed {package}"

    return node_detected_package


def detect_installed_packages_version_hosts(
        cluster: KubernetesCluster, hosts_to_packages: Mapping[str, Union[str, List[str]]]
) -> Dict[str, Dict[str, List[str]]]:
    """
    Detect grouped packages versions for specified list of packages for each remote host.

    :param cluster: KubernetesCluster instance
    :param hosts_to_packages: Remote hosts with list of packages to detect versions.
    :return: Dictionary with grouped versions for each queried package, pointing to list of hosts,
        e.g. {"foo" -> {"foo-1": [host1, host2]}, "bar" -> {"bar-1": [host1], "bar-2": [host2]}}
    """
    hosts_to_packages_dedup = {}
    for host, packages_list in hosts_to_packages.items():
        if isinstance(packages_list, str):
            packages_list = [packages_list]
        # deduplicate
        hosts_to_packages_dedup[host] = list(set(packages_list))

    collector = CollectorCallback(cluster)
    with cluster.make_group(hosts_to_packages).new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            for package in hosts_to_packages_dedup[node.get_host()]:
                _detect_installed_package_version(node, package, collector)

    results: Dict[str, Dict[str, List]] = {}

    for host, multiple_results in collector.results.items():
        for i, result in enumerate(multiple_results):
            package = hosts_to_packages_dedup[host][i]
            node_detected_package = _parse_node_detected_package(result, package)
            results.setdefault(package, {}).setdefault(node_detected_package, []).append(host)

    return results


def get_package_version_separator(os_family: str) -> str:
    return '=' if os_family == 'debian' else '-'


def get_package_name(os_family: str, package: str) -> str:
    """
    Return the pure package name, without any part of version
    """

    package_name = ""

    if package:
        if os_family in ["rhel", "rhel8", "rhel9"]:
            # regexp is needed to split package and its version, the pattern start with '-' then should be number or '*'
            package_name = re.split(r'-[\d,\*]', package)[0]
        else:
            # in ubuntu it is much easier to parse package name
            package_name = package.split("=")[0]

    return package_name



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
from abc import ABC, abstractmethod
from functools import total_ordering

from kubemarine.core.action import Action


@total_ordering
class _Priority(enum.Enum):
    INVENTORY_ONLY = 0
    """
    The patch may change the inventory.
    The patch can run only LIGHT enrichment, and connect to nodes for read-only aims.
    """

    SOFTWARE_UPGRADE = 1
    "This is a service patch that should be instantiated only automatically by migrate_kubemarine.py"

    REGULAR = 2
    """
    The patch can access and make some operations on the cluster.
    Changes in the inventory are possible, but they should not affect the software upgrade procedure.
    """

    def __lt__(self, other: '_Priority') -> bool:
        return self.value < other.value


class Patch(ABC):
    def __init__(self, identifier: str):
        self.identifier = identifier

    @abstractmethod
    def priority(self) -> _Priority:
        pass

    @property
    @abstractmethod
    def action(self) -> Action:
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        pass


class InventoryOnlyPatch(Patch, ABC):
    """
    The patch may change the inventory.
    Calling DynamicResources.cluster(EnrichmentStage.LIGHT) is allowed to connect to nodes for read-only aims.

    Patches if this type are executed first.
    """

    def priority(self) -> _Priority:
        return _Priority.INVENTORY_ONLY


class _SoftwareUpgradePatch(Patch, ABC):
    """This is a service patch that should be extended only by predefined set of classes inside migrate_kubemarine.py"""

    def priority(self) -> _Priority:
        return _Priority.SOFTWARE_UPGRADE


class RegularPatch(Patch, ABC):
    """
    The patch can access and make some operations on the cluster.

    Changes in the inventory are possible,
    but the cluster should be `DynamicResources.reset_cluster()` for the changes to take effect.
    The changes in the inventory should also not affect the software upgrade procedure.

    Patches if this type are executed last.
    """

    def priority(self) -> _Priority:
        return _Priority.REGULAR



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
from io import StringIO
from typing import Dict, List, Tuple, Optional

import toml
import yaml

from kubemarine import system, packages, jinja
from kubemarine.core import utils, static, errors
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup, RunnersGroupResult, CollectorCallback
from kubemarine.core.yaml_merger import default_merger


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    containerd_config = inventory['services']['cri']['containerdConfig']

    path = 'plugins."io.containerd.grpc.v1.cri"'
    kubernetes_version = inventory['services']['kubeadm']['kubernetesVersion']
    containerd_config[path].setdefault('sandbox_image', get_default_sandbox_image(inventory, kubernetes_version))

    # Check if field for new and old configuration formats are presented
    old_format_result, old_format_field = contains_old_format_properties(inventory)
    new_format_result, new_format_field = contains_new_format_properties(inventory)

    # Fail if fields for both formats are presented
    if old_format_result and new_format_result:
        raise errors.FailException(f"Invalid containerd configuration: "
                                   f"{old_format_field} can't be set when {new_format_field} is provided")

    # If no fields for old format, enrich config_path default value to new format
    if not old_format_result and 'containerdRegistriesConfig' in inventory['services']['cri']:
        containerd_config.setdefault('plugins."io.containerd.grpc.v1.cri".registry', {})\
            .setdefault('config_path', '/etc/containerd/certs.d')


def contains_old_format_properties(inventory: dict) -> Tuple[bool, Optional[str]]:
    config_toml = get_config_as_toml(inventory.get("services", {}).get("cri", {}).get('containerdConfig', {}))
    if "mirrors" in config_toml.get("plugins", {}).get("io.containerd.grpc.v1.cri", {}).get("registry", {}):
        return True, 'mirrors for "io.containerd.grpc.v1.cri" plugin in services.cri.containerdConfig'
    for _, config in config_toml.get("plugins", {}).get("io.containerd.grpc.v1.cri", {})\
            .get("registry", {}).get("configs", {}).items():
        if "tls" in config:
            return True, 'configs.tls for "io.containerd.grpc.v1.cri" plugin in services.cri.containerdConfig'
    return False, None


def contains_new_format_properties(inventory: dict) -> Tuple[bool, Optional[str]]:
    config_toml = get_config_as_toml(inventory.get("services", {}).get("cri", {}).get('containerdConfig', {}))
    if "config_path" in config_toml.get("plugins", {}).get("io.containerd.grpc.v1.cri", {}).get("registry", {}):
        return True, 'config_path for "io.containerd.grpc.v1.cri" plugin in services.cri.containerdConfig'
    if inventory.get('services', {}).get('cri', {}).get('containerdRegistriesConfig', {}):
        return True, 'services.cri.containerdRegistriesConfig'
    return False, None


def get_default_sandbox_image(inventory: dict, kubernetes_version: str) -> str:
    image_repository = inventory['services']['kubeadm']['imageRepository']
    pause_version = static.GLOBALS['compatibility_map']['software']['pause'][kubernetes_version]['version']
    return f"{image_repository}/pause:{pause_version}"


def get_sandbox_image(cri_config: dict) -> Optional[str]:
    sandbox_image: Optional[str] = cri_config.get('containerdConfig', {})\
        .get('plugins."io.containerd.grpc.v1.cri"', {}).get('sandbox_image')

    return sandbox_image


def get_sandbox_image_upgrade_plan(cluster: KubernetesCluster) -> List[Tuple[str, dict]]:
    context = cluster.context
    procedure_inventory = cluster.procedure_inventory
    upgrade_plan = []

    if context["initial_procedure"] == "upgrade":
        kubernetes_version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
        for i, v in enumerate(procedure_inventory['upgrade_plan'][context['upgrade_step']:]):
            # Take the target (probably) compiled version and the remained not yet compiled
            version = kubernetes_version if i == 0 else v
            upgrade_config = procedure_inventory.get(v, {}).get('cri', {})
            upgrade_plan.append((version, upgrade_config))

    return upgrade_plan


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def enrich_upgrade_inventory(cluster: KubernetesCluster) -> None:
    upgrade_plan = get_sandbox_image_upgrade_plan(cluster)
    upgrade_config = {} if not upgrade_plan else upgrade_plan[0][1]

    if upgrade_config:
        default_merger.merge(cluster.inventory.setdefault("services", {}).setdefault("cri", {}), upgrade_config)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def verify_upgrade_inventory(cluster: KubernetesCluster) -> None:
    upgrade_plan = get_sandbox_image_upgrade_plan(cluster)
    if not upgrade_plan:
        return

    previous_version = cluster.previous_inventory['services']['kubeadm']['kubernetesVersion']
    sandbox_image = get_sandbox_image(cluster.previous_raw_inventory.get('services', {}).get('cri', {}))
    for version, upgrade_config in upgrade_plan:
        if version != '' and jinja.is_template(version):
            break

        upgrade_sandbox_image = get_sandbox_image(upgrade_config)
        if sandbox_image and not upgrade_sandbox_image:
            raise errors.KME("KME0013",
                             previous_version_spec=f' for version {previous_version}',
                             next_version_spec=f' for version {version}')

        sandbox_image = upgrade_sandbox_image
        previous_version = version


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def calculate_sandbox_image_upgrade_required(cluster: KubernetesCluster) -> None:
    upgrade_required = (get_sandbox_image(cluster.previous_inventory['services']['cri'])
                        != get_sandbox_image(cluster.inventory['services']['cri']))

    cluster.context.setdefault("upgrade", {}).setdefault('required', {})['containerdConfig'] = upgrade_required


def fetch_containerd_config(group: NodeGroup) -> Tuple[Dict[str, dict], Dict[str, dict]]:
    cluster = group.cluster
    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            config_location = cluster.get_package_association_for_node(node.get_host(),
                                                                       'containerd', 'config_location')
            node.sudo(f'cat {config_location}', callback=collector)

    containerd_config = {host: toml.loads(config_string.stdout)
                         for host, config_string in collector.result.items()}

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            config_path = containerd_config[node.get_host()]\
                .get('plugins', {})\
                .get('io.containerd.grpc.v1.cri', {}).get('registry', {}).get('config_path')
            if config_path:
                node.sudo(f'ls {config_path}', callback=collector)

    containerd_registries = {host: registries.stdout.split('\n')[:-1]
                             for host, registries in collector.result.items()}

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            config_path = containerd_config[node.get_host()] \
                .get('plugins', {}) \
                .get('io.containerd.grpc.v1.cri', {}).get('registry', {}).get('config_path')
            for registry in containerd_registries.get(node.get_host(), {}):
                node.sudo(f'cat {config_path}/{registry}/hosts.toml', callback=collector)

    containerd_reg_config = {host: {containerd_registries[host][i]: toml.loads(reg_host.stdout)
                                    for i, reg_host in enumerate(reg_hosts)}
                             for host, reg_hosts in collector.results.items()}

    return containerd_config, containerd_reg_config


def install(group: NodeGroup) -> RunnersGroupResult:
    collector = CollectorCallback(group.cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            os_specific_associations = exe.cluster.get_associations_for_node(node.get_host(), 'containerd')

            exe.cluster.log.debug("Installing latest containerd on %s node" % node.get_node_name())
            # always install latest available containerd
            packages.install(node, include=os_specific_associations['package_name'], pty=True, callback=collector)

            # remove previous config.toml to avoid problems in case when previous config was broken
            node.sudo("rm -f %s && sudo systemctl restart %s"
                      % (os_specific_associations['config_location'],
                         os_specific_associations['service_name']),
                      callback=collector)

            system.enable_service(node, os_specific_associations['service_name'],
                                  now=True, callback=collector)
    return collector.result


def get_config_as_toml(config: dict) -> dict:
    config_string = ''
    # double loop is used to make sure that no "simple" `key: value` pairs are accidentally assigned to sections
    for key, value in config.items():
        # first we process all "simple" `key: value` pairs
        if not isinstance(value, dict):
            config_string += f"{toml.dumps({key: value})}"
    for key, value in config.items():
        # next we process all "complex" `key: dict_value` pairs, representing named sections
        if isinstance(value, dict):
            config_string += f"\n[{key}]\n{toml.dumps(value)}"

    return toml.loads(config_string)


def configure_ctr_flags(group: NodeGroup) -> None:
    cluster = group.cluster
    log = cluster.log
    config_toml = get_config_as_toml(cluster.inventory.get("services", {}).get("cri", {}).get('containerdConfig', {}))
    # Calculate ctr options for image pull
    registry = config_toml.get('plugins', {}).get('io.containerd.grpc.v1.cri', {}).get('registry', {})
    config_path = registry.get('config_path')
    ctr_pull_options_str = ""
    for registry_name in set().union(registry.get('mirrors', {}).keys(), registry.get('configs', {}).keys()):
        options = []
        # Add plain-http flag for http endpoint
        if any("http://" in endpoint for endpoint in
               registry.get('mirrors', {}).get(registry_name, {}).get('endpoint', [])):
            options.append('--plain-http')
        # Add skip-verify flag in case of insecure tls connection
        if registry.get('configs', {}).get(registry_name, {}).get('tls', {}).get('insecure_skip_verify', False):
            options.append('--skip-verify')
        # Add user flag if authorization required
        registry_auth = registry.get('configs', {}).get(registry_name, {}).get('auth', {})
        if registry_auth.get('auth'):
            options.append(f'--user {base64.b64decode(registry_auth["auth"]).decode("utf-8")}')
        elif registry_auth.get('username'):
            options.append(f'--user {registry_auth["username"]}' +
                           f':{registry_auth["password"]}' if registry_auth.get("password") else '')
        ctr_pull_options_str += f'{registry_name}={" ".join(options)}\n'

    # Add hosts-dir, if it's presented
    if config_path:
        ctr_pull_options_str += f'*=--hosts-dir {config_path}\n'

    # Save ctr pull options
    log.debug("Uploading ctr flags configuration...")
    group.put(StringIO(ctr_pull_options_str), "/etc/ctr/kubemarine_ctr_flags.conf", backup=True, sudo=True, mkdir=True)
    group.sudo("chmod 600 /etc/ctr/kubemarine_ctr_flags.conf")


def configure_crictl(group: NodeGroup) -> None:
    cluster = group.cluster
    cluster.log.debug("Uploading crictl configuration for containerd...")
    crictl_config = yaml.dump({"runtime-endpoint": "unix:///run/containerd/containerd.sock"})
    utils.dump_file(cluster, crictl_config, 'crictl.yaml')
    group.put(StringIO(crictl_config), '/etc/crictl.yaml', backup=True, sudo=True)


def get_config_path(inventory: dict) -> Optional[str]:
    config_path: Optional[str] = inventory.get('services', {}).get('cri', {}).get('containerdConfig', {}) \
        .get('plugins."io.containerd.grpc.v1.cri".registry', {}).get('config_path')
    return config_path


def configure_containerd(group: NodeGroup) -> RunnersGroupResult:
    cluster = group.cluster
    log = cluster.log

    # Dump containerd configuration
    config_toml = get_config_as_toml(cluster.inventory.get("services", {}).get("cri", {}).get('containerdConfig', {}))
    config_string = toml.dumps(config_toml)
    utils.dump_file(cluster, config_string, 'containerd-config.toml')
    config_path = config_toml.get('plugins', {}).get('io.containerd.grpc.v1.cri', {})\
        .get('registry', {}).get('config_path')

    # Dump registries configuration
    registries_config = {}
    for registry, host_config in cluster.inventory.get("services", {}).get("cri", {})\
            .get('containerdRegistriesConfig', {}).items():
        registry_host_toml = get_config_as_toml(host_config)
        registries_config[registry] = toml.dumps(registry_host_toml)
        utils.dump_file(cluster, registries_config[registry], f"registries/{registry}/hosts.toml")

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            os_specific_associations = exe.cluster.get_associations_for_node(node.get_host(), 'containerd')
            log.debug("Uploading containerd configuration to %s node..." % node.get_node_name())
            node.put(StringIO(config_string), os_specific_associations['config_location'],
                     backup=True, sudo=True, mkdir=True)

            if config_path:
                log.debug("Uploading containerd registries configuration to %s on %s node..." %
                          (config_path, node.get_node_name()))
                node.sudo(f'mkdir -p {config_path} && sudo rm -fr {config_path}/*')
                for registry, host_config in registries_config.items():
                    node.put(StringIO(host_config), f'{config_path}/{registry}/hosts.toml',
                             backup=True, sudo=True, mkdir=True)

            log.debug("Restarting Containerd on %s node..." % node.get_node_name())
            node.sudo(
                f"chmod 600 {os_specific_associations['config_location']} && "
                f"sudo systemctl restart {os_specific_associations['service_name']} && "
                f"systemctl status {os_specific_associations['service_name']}", callback=collector)
    return collector.result


def configure(group: NodeGroup) -> RunnersGroupResult:
    configure_crictl(group)
    configure_ctr_flags(group)
    return configure_containerd(group)


def prune(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo('crictl rm -fa; '
                      'sudo crictl rmp -fa; '
                      'sudo crictl rmi -a; '
                      'sudo ctr content ls -q | xargs -r sudo ctr content rm', warn=True)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
from typing import List, Dict, Optional, Tuple

from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, CollectorCallback


# the methods requires etcdctl.sh to be installed on all active control-plane nodes during thirdparties task.

def remove_members(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    initial_control_planes = cluster.previous_nodes["control-plane"]
    managing_control_plane = cluster.get_unchanged_nodes().having_roles(['control-plane']).get_any_member()

    log.verbose(f"etcd will be managed using {managing_control_plane.get_node_name()}.")
    output = managing_control_plane.sudo("etcdctl member list").get_simple_out().splitlines()

    etcd_members = {}
    for line in output:
        params = [p.strip() for p in line.split(sep=',')]
        # 6 is expected number of comma-separated parameters of an etcd member
        if len(params) == 6:
            etcd_members[params[2]] = params[0]
        else:
            log.warning("Unexpected line in 'etcdctl member list' output: " + line)

    log.verbose(f"Found etcd members {list(etcd_members.keys())}")
    unexpected_members = etcd_members.keys() - set(initial_control_planes.get_nodes_names())
    if unexpected_members:
        log.warning(f"Found unexpected etcd members {list(unexpected_members)}")

    for node_name in group.get_nodes_names():
        if node_name in etcd_members:
            command = "etcdctl member remove " + etcd_members[node_name]
            log.verbose(f"Removing found etcd member {node_name}...")
            managing_control_plane.sudo(command, pty=True)
        else:
            log.verbose(f"Skipping {node_name} as it is not among etcd members.")


def wait_for_health(cluster: KubernetesCluster, node: NodeGroup) -> List[Dict]:
    """
    The method checks etcd endpoints health until all endpoints are healthy or retries are exhausted
    if all member are healthy the method checks the leader.
    """
    log = cluster.log
    timeout = cluster.globals['etcd']['health']['timeout']
    retries = cluster.globals['etcd']['health']['retries']

    is_healthy = False
    while retries > 0:
        start_time = time.time()
        is_healthy = _is_healthy(cluster, node)
        end_time = time.time()
        sudo_time = int(end_time - start_time)

        if is_healthy:
            log.debug('All ETCD members are healthy!')
            break

        log.debug('Wait for ETCD cluster is not healthy!')
        if sudo_time < timeout:
            time.sleep(timeout - sudo_time)
        retries -= 1

    if is_healthy:
        _, etcd_status_list = execute_endpoints_command(cluster, node, 'status', warn=False)
        elected_leader: Optional[int] = None
        for item in etcd_status_list:
            leader: Optional[int] = item.get('status', {}).get('leader')
            if leader is None:
                raise Exception('ETCD member "%s" do not have leader' % item.get('endpoint'))
            if elected_leader is None:
                elected_leader = leader
            elif elected_leader != leader:
                raise Exception('ETCD leaders are not the same')
        log.debug('Leader "%s" elected' % elected_leader)
    else:
        raise Exception('ETCD cluster is still not healthy!')
      
    log.verbose('ETCD cluster is healthy!')

    return etcd_status_list


def _is_healthy(cluster: KubernetesCluster, node: NodeGroup) -> bool:
    is_healthy, etcd_health_list = execute_endpoints_command(cluster, node, 'health', warn=True)
    if not is_healthy:
        return False

    health = sum(1 for etcd_health in etcd_health_list if etcd_health.get('health'))
    if health != cluster.nodes['control-plane'].nodes_amount():
        return False

    return True


def execute_endpoints_command(cluster: KubernetesCluster, node: NodeGroup, command: str,
                              *, warn: bool) -> Tuple[bool, List[dict]]:
    logger = cluster.log
    tmp_path = utils.get_remote_tmp_path()
    host = node.get_host()

    defer = node.new_defer()
    collector = CollectorCallback(cluster)

    # Separate stdout and stderr using extra temporary file
    defer.sudo(f'etcdctl endpoint {command} --cluster -w json > {tmp_path}',
               pty=True, warn=warn, callback=collector)
    defer.sudo(f'cat {tmp_path}', pty=True, warn=warn, callback=collector)
    defer.sudo(f'rm -f {tmp_path}', pty=True, warn=warn, callback=collector)

    defer.flush()
    if collector.result.is_any_failed():
        logger.verbose(collector.result)
        return False, []

    endpoints_raw = collector.results[host][1].stdout

    cluster.log.verbose(endpoints_raw)
    endpoints_list: List[dict] = json.loads(endpoints_raw.lower().rstrip('\n'))

    return True, endpoints_list



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Iterator, Mapping, List


class Environ(Mapping[str, str]):
    """
    Read-only view of os.environ.
    """

    def __getitem__(self, name: str) -> str:
        # check presence of the variable and throw KeyError if necessary
        return os.environ[name]

    def __len__(self) -> int:
        return len(os.environ)

    def __iter__(self) -> Iterator[str]:
        return iter(os.environ)

    __slots__: List[str] = []



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import io
from typing import Union, Optional, List, Dict

from kubemarine.core import utils
from kubemarine.core.executor import RunnersResult, Token, Callback
from kubemarine.core.group import (
    NodeGroup, RunnersGroupResult, AbstractGroup, RunResult, DeferredGroup, GROUP_RUN_TYPE
)


def ls_repofiles(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo('ls -la /etc/yum.repos.d')


def backup_repo(group: NodeGroup) -> Optional[RunnersGroupResult]:
    if not group.cluster.inventory['services']['packages']['package_manager']['replace-repositories']:
        group.cluster.log.debug("Skipped - repos replacement disabled in configuration")
        return None
    # all files in directory will be renamed: xxx.repo -> xxx.repo.bak
    # if there already any files with ".bak" extension, they should not be renamed to ".bak.bak"!
    return group.sudo("find /etc/yum.repos.d/ -type f -name '*.repo' | "
                      "sudo xargs -t -iNAME mv -bf NAME NAME.bak")


def add_repo(group: NodeGroup, repo_data: Union[List[str], Dict[str, dict], str]) -> RunnersGroupResult:
    create_repo_file(group, repo_data, get_repo_file_name())
    return group.sudo('yum clean all && sudo yum updateinfo -d1', pty=True)


def get_repo_file_name() -> str:
    return '/etc/yum.repos.d/predefined.repo'


def create_repo_file(group: AbstractGroup[RunResult],
                     repo_data: Union[List[str], Dict[str, dict], str],
                     repo_file: str) -> None:
    # if repo_data is dict, then convert it to string with config inside
    if isinstance(repo_data, dict):
        config = configparser.ConfigParser()
        for repo_id, data in repo_data.items():
            config[repo_id] = data
        repo_data_stream = io.StringIO()
        config.write(repo_data_stream)
    elif isinstance(repo_data, list):
        raise Exception("Not supported repositories format for yum package manager")
    else:
        repo_data_stream = io.StringIO(utils.read_external(repo_data))

    group.put(repo_data_stream, repo_file, sudo=True)


def clean(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo("yum clean all", pty=True)


def get_install_cmd(include: Union[str, List[str]], exclude: Union[str, List[str]] = None) -> str:
    if isinstance(include, list):
        include = ' '.join(include)
    command = 'yum install -y -d1 --color=never %s' % include

    if exclude is not None:
        if isinstance(exclude, list):
            exclude = ','.join(exclude)
        command += ' --exclude=%s' % exclude
    command += f"; PACKAGES=$(rpm -q {include}); " \
               f"if [ $? != 0 ]; then echo \"$PACKAGES\" | grep 'is not installed'; " \
               f"echo \"Failed to check version for some packages. " \
               f"Make sure packages are not already installed with higher versions. " \
               f"Also, make sure user-defined packages have rpm-compatible names. \"; exit 1; fi "

    return command


def install(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None,
            pty: bool = False, callback: Callback = None) -> GROUP_RUN_TYPE:
    if include is None:
        raise Exception('You must specify included packages to install')

    command = get_install_cmd(include, exclude)

    return group.sudo(command, pty=pty, callback=callback)


def remove(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
           exclude: Union[str, List[str]] = None,
           warn: bool = False, hide: bool = True, pty: bool = False) -> GROUP_RUN_TYPE:
    if include is None:
        raise Exception('You must specify included packages to remove')

    if isinstance(include, list):
        include = ' '.join(include)
    command = 'yum remove -y -d1 --color=never %s' % include

    if exclude is not None:
        if isinstance(exclude, list):
            exclude = ','.join(exclude)
        command += ' --exclude=%s' % exclude

    return group.sudo(command, warn=warn, hide=hide, pty=pty)


def upgrade(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None,
            pty: bool = False) -> GROUP_RUN_TYPE:
    if include is None:
        raise Exception('You must specify included packages to upgrade')

    if isinstance(include, list):
        include = ' '.join(include)
    command = 'yum upgrade -y -d1 --color=never %s' % include

    if exclude is not None:
        if isinstance(exclude, list):
            exclude = ','.join(exclude)
        command += ' --exclude=%s' % exclude

    return group.sudo(command, pty=pty)


def no_changes_found(action: str, result: RunnersResult) -> bool:
    if action not in ('install', 'upgrade', 'remove'):
        raise Exception(f"Unknown action {action}")

    output = result.stdout.rstrip('\n')
    if "Nothing to do." in output:
        return True
    # Remove next checks after `rhel` (version 7) OS family support stop.
    if action == 'install' and all(
            'already installed and latest version' in line
            for line in output.split('\n')):
        return True
    if action == 'upgrade' and not output or all(
            'No Match for argument' in line
            for line in output.split('\n')):
        return True
    if action == 'remove' and all(
            'No Match for argument' in line
            for line in output.split('\n')):
        return True

    return False


def search(group: DeferredGroup, package: str, callback: Callback = None) -> Token:
    if package is None:
        raise Exception('You must specify package to search')
    command = 'yum list -d1 %s' % package

    return group.sudo(command, pty=True, warn=True, callback=callback)
